@@ -87,7 +87,7 @@ KWD::Window::Window (WId  parentId,
     mMapped (false),
     mPendingMap (0),
     mPendingConfigure (0),
-    mProcessKiller (0),
+    mProcessKiller (this),
     mKeys (this),
     mResizeOpAction (0),
     mMoveOpAction (0),
@@ -117,18 +117,28 @@ KWD::Window::Window (WId  parentId,
 					 KWindowSystem::NETWM |
 					 KWindowSystem::WMHints );
 
-	    if (mIcon.isNull())
-		mIcon = KWindowSystem::icon (mClientId, 16, 16, true,
+	    mMiniIcon = KWindowSystem::icon (mClientId, 16, 16, true,
 					     KWindowSystem::NETWM |
 					     KWindowSystem::WMHints );
 
+	    if (mIcon.isNull ())
+	    {
+		mIcon = KWindowSystem::icon (mClientId, 32, 32, true,
+					     KWindowSystem::ClassHint |
+					     KWindowSystem::XApp );
+		mMiniIcon = KWindowSystem::icon (mClientId, 16, 16, true,
+						 KWindowSystem::ClassHint |
+						 KWindowSystem::XApp );
+	    }
+
 	    mOpacity = readPropertyShort (mClientId, Atoms::netWmWindowOpacity,
 					  0xffff);
 	}
 	else
 	{
-	    mIcon = QIcon ();
-	    mName = QString ("");
+	    mIcon     = QPixmap ();
+	    mMiniIcon = QPixmap ();
+	    mName     = QString ("");
 	}
 
 	updateFrame (frame);
@@ -139,7 +149,8 @@ KWD::Window::Window (WId  parentId,
     }
     else
     {
-	mIcon = QIcon ();
+	mIcon     = QPixmap ();
+	mMiniIcon = QPixmap ();
 	mName = QString ("");
 	mGeometry = QRect (50, 50, 30, 1);
     }
@@ -172,8 +183,16 @@ KWD::Window::~Window (void)
     if (mPopup)
 	delete mPopup;
 
-    if (mProcessKiller)
-	delete mProcessKiller;
+    if (mProcessKiller.state () == QProcess::Running)
+    {
+	mProcessKiller.terminate ();
+	mProcessKiller.waitForFinished (10000);
+	if (mProcessKiller.state () == QProcess::Running)
+	{
+	    mProcessKiller.kill ();
+	    mProcessKiller.waitForFinished (5000);
+	}
+    }
 }
 
 bool
@@ -335,7 +354,9 @@ KWD::Window::isResizable (void) const
 QIcon
 KWD::Window::icon (void) const
 {
-    return mIcon;
+    QIcon icon (mIcon);
+    icon.addPixmap (mMiniIcon);
+    return icon;
 }
 
 QString
@@ -1147,7 +1168,7 @@ KWD::Window::setMask (const QRegion &reg, int)
 	mUniqueVertShape = uniqueVertShape;
 
 	if (mPixmap)
-	    updateShadow ();
+	    QTimer::singleShot (0, this, SLOT (updateShadow ()));
     }
 }
 
@@ -1288,8 +1309,6 @@ KWD::Window::rebindPixmap (void)
 				     xformat, 0, NULL);
 
     updateShadow ();
-
-    mDecor->widget ()->update ();
 }
 
 bool
@@ -1830,10 +1849,20 @@ KWD::Window::updateIcons (void)
 				 KWindowSystem::NETWM |
 				 KWindowSystem::WMHints);
 
-    if (mIcon.isNull())
-	mIcon = KWindowSystem::icon (mClientId, 16, 16, true,
+    mMiniIcon = KWindowSystem::icon (mClientId, 16, 16, true,
 				     KWindowSystem::NETWM |
 				     KWindowSystem::WMHints);
+
+    if (mIcon.isNull ())
+    {
+	mIcon = KWindowSystem::icon (mClientId, 32, 32, true,
+				     KWindowSystem::ClassHint |
+				     KWindowSystem::XApp );
+	mMiniIcon = KWindowSystem::icon (mClientId, 16, 16, true,
+					 KWindowSystem::ClassHint |
+					 KWindowSystem::XApp );
+    }
+
     mDecor->iconChange ();
 }
 
@@ -1881,6 +1910,8 @@ KWD::Window::moveWindow (QMouseEvent *qme)
 					       qme->globalX (),
 					       qme->globalY (),
 					       direction);
+    mFakeRelease = true;
+
 }
 
 #define OPACITY_STEP (0xffff / 10)
@@ -2025,15 +2056,17 @@ KWD::Window::processDamage (void)
     {
 	r2.translate (-xOff, -yOff);
 
-	decor_blend_top_border_picture (QX11Info::display(),
-					&mContext,
-					mPicture,
-					xOff, xOff,
-					mTexturePicture,
-					&mLayout,
-					r2.handle (),
-					(unsigned short) (alpha * 0xffff),
-					shade_alpha);
+	decor_blend_border_picture (QX11Info::display(),
+				    &mContext,
+				    mPicture,
+				    xOff, xOff,
+				    mTexturePicture,
+				    &mLayout,
+				    BORDER_TOP,
+				    r2.handle (),
+				    (unsigned short) (alpha * 0xffff),
+				    shade_alpha,
+				    TRUE);
     }
 
     xOff = 0;
@@ -2046,15 +2079,17 @@ KWD::Window::processDamage (void)
     {
 	r2.translate (-xOff, -yOff);
 
-	decor_blend_bottom_border_picture (QX11Info::display(),
-					   &mContext,
-					   mPicture,
-					   xOff, yOff,
-					   mTexturePicture,
-					   &mLayout,
-					   r2.handle (),
-					   (unsigned short) (alpha * 0xffff),
-					   shade_alpha);
+	decor_blend_border_picture (QX11Info::display(),
+				    &mContext,
+				    mPicture,
+				    xOff, yOff,
+				    mTexturePicture,
+				    &mLayout,
+				    BORDER_BOTTOM,
+				    r2.handle (),
+				    (unsigned short) (alpha * 0xffff),
+				    shade_alpha,
+				    TRUE);
     }
 
     xOff = 0;
@@ -2067,15 +2102,17 @@ KWD::Window::processDamage (void)
     {
 	r2.translate (-xOff, -yOff);
 
-	decor_blend_left_border_picture (QX11Info::display(),
-					 &mContext,
-					 mPicture,
-					 xOff, yOff,
-					 mTexturePicture,
-					 &mLayout,
-					 r2.handle (),
-					 (unsigned short) (alpha * 0xffff),
-					 shade_alpha);
+	decor_blend_border_picture (QX11Info::display(),
+				    &mContext,
+				    mPicture,
+				    xOff, yOff,
+				    mTexturePicture,
+				    &mLayout,
+				    BORDER_LEFT,
+				    r2.handle (),
+				    (unsigned short) (alpha * 0xffff),
+				    shade_alpha,
+				    TRUE);
     }
 
     xOff = mContext.extents.left + mGeometry.width ();
@@ -2088,15 +2125,17 @@ KWD::Window::processDamage (void)
     {
 	r2.translate (-xOff, -yOff);
 
-	decor_blend_right_border_picture (QX11Info::display(),
-					  &mContext,
-					  mPicture,
-					  xOff, yOff,
-					  mTexturePicture,
-					  &mLayout,
-					  r2.handle (),
-					  (unsigned short) (alpha * 0xffff),
-					  shade_alpha);
+	decor_blend_border_picture (QX11Info::display(),
+				    &mContext,
+				    mPicture,
+				    xOff, yOff,
+				    mTexturePicture,
+				    &mLayout,
+				    BORDER_RIGHT,
+				    r2.handle (),
+				    (unsigned short) (alpha * 0xffff),
+				    shade_alpha,
+				    TRUE);
     }
 
     mDamage = QRegion ();
@@ -2116,16 +2155,6 @@ KWD::Window::processDamage (void)
 	updateProperty ();
 }
 
-void
-KWD::Window::handleProcessKillerExited (void)
-{
-    if (mProcessKiller)
-    {
-	delete mProcessKiller;
-	mProcessKiller = NULL;
-    }
-}
-
 void
 KWD::Window::showKillProcessDialog (Time timestamp)
 {
@@ -2138,7 +2167,7 @@ KWD::Window::showKillProcessDialog (Time timestamp)
     pid_t	     pid;
     char	     buf[257];
 
-    if (mProcessKiller)
+    if (mProcessKiller.state () == QProcess::Running)
 	return;
 
     clientMachine = kWinInfo.clientMachine ();
@@ -2151,25 +2180,21 @@ KWD::Window::showKillProcessDialog (Time timestamp)
 	    clientMachine = "localhost";
     }
 
-    mProcessKiller = new QProcess (this);
-
-    connect (mProcessKiller, SIGNAL (finished (int, QProcess::ExitStatus)),
-	     SLOT (handleProcessKillerExited ()));
-    connect (mProcessKiller, SIGNAL (error (QProcess::ProcessError)),
-	     SLOT (handleProcessKillerExited ()));
-    
-
-    mProcessKiller->start (KStandardDirs::findExe ("kwin_killer_helper"),
+    mProcessKiller.start (KStandardDirs::findExe ("kwin_killer_helper"),
 	QStringList () << "--pid" << QByteArray ().setNum (pid) <<
 	"--hostname" << clientMachine <<
 	"--windowname" << mName.toUtf8 () <<
 	"--applicationname" << resourceClass <<
 	"--wid" << QByteArray ().setNum ((unsigned int) mClientId) <<
-	"--timestamp" << QByteArray ().setNum ((unsigned int) timestamp));
+	"--timestamp" << QByteArray ().setNum ((unsigned int) timestamp),
+	QIODevice::NotOpen);
 }
 
 void
 KWD::Window::hideKillProcessDialog (void)
 {
-    handleProcessKillerExited ();
+    if (mProcessKiller.state () == QProcess::Running)
+    {
+	mProcessKiller.terminate ();
+    }
 }
