@@ -27,11 +27,13 @@
 #include <fixx11h.h>
 #include <kapplication.h>
 #include <kcmdlineargs.h>
+#include <klocale.h>
 
 #include "decorator.h"
 
 static const KCmdLineOptions options[] = {
     { "replace", "Replace existing window decorator", 0 },
+    { "sm-disable", "Disable connection to session manager", 0 },
     { "opacity <value>", "Decoration opacity", "0.75" },
     { "no-opacity-shade", "No decoration opacity shading", 0 },
     { "active-opacity <value>", "Active decoration opacity", "1.0" },
@@ -50,6 +52,8 @@ main (int argc, char **argv)
     int		   event, error;
     Time	   timestamp;
 
+    KLocale::setMainCatalogue("kwin");
+
     KCmdLineArgs::init (argc, argv,
 			"kde-window-decorator",
 			"KWD",
@@ -83,6 +87,9 @@ main (int argc, char **argv)
 
     app = new KWD::Decorator ();
 
+    if (args->isSet ("sm-disable"))
+	app->disableSessionManagement ();
+
     if (!XDamageQueryExtension (qt_xdisplay (), &event, &error))
     {
 	fprintf (stderr,
