@@ -155,6 +155,7 @@ typedef struct _BlurScreen {
     int    height;
 
     GLuint program;
+    int    maxTemp;
     GLuint fbo;
     Bool   fboStatus;
 
@@ -1125,10 +1126,13 @@ getDstBlurFragmentFunction (CompScreen  *s,
     if (data)
     {
 	static char *temp[] = { "fCoord", "mask", "sum", "dst" };
-	int	    i, handle = 0;
+	int	    i, j, handle = 0;
 	char	    str[1024];
 	int	    saturation = bs->opt[BLUR_SCREEN_OPTION_SATURATION].value.i;
 	Bool	    ok = TRUE;
+	int	    numIndirect;
+	int	    numIndirectOp;
+	int	    base, end, ITCbase;
 
 	for (i = 0; i < sizeof (temp) / sizeof (temp[0]); i++)
 	    ok &= addTempHeaderOpToFunctionData (data, temp[i]);
@@ -1184,64 +1188,53 @@ getDstBlurFragmentFunction (CompScreen  *s,
 	    ok &= addDataOpToFunctionData (data, str);
 	} break;
 	case BLUR_FILTER_GAUSSIAN: {
-	    static char *filterTemp[] = {
-		"tCoord", "pix"
-	    };
 
-	    for (i = 0; i < sizeof (filterTemp) / sizeof (filterTemp[0]); i++)
-		ok &= addTempHeaderOpToFunctionData (data, filterTemp[i]);
+	    /* try to use only half of the available temporaries to keep
+	       other plugins working */
+	    if ((bs->maxTemp / 2) - 4 >
+		 (bs->numTexop + (bs->numTexop - numITC)) * 2)
+	    {
+		numIndirect   = 1;
+		numIndirectOp = bs->numTexop;
+	    }
+	    else
+	    {
+		i = MAX(((bs->maxTemp / 2) - 4) / 4, 1);
+		numIndirect = ceil ((float)bs->numTexop / (float)i);
+		numIndirectOp = ceil ((float)bs->numTexop / (float)numIndirect);
+	    }
 
-	    for (i = 0; i < bs->numTexop * 2; i++)
+	    /* we need to define all coordinate temporaries if we have
+	       multiple indirection steps */
+	    j = (numIndirect > 1) ? 0 : numITC;
+
+	    for (i = 0; i < numIndirectOp * 2; i++)
 	    {
 		snprintf (str, 1024, "pix_%d", i);
 		ok &= addTempHeaderOpToFunctionData (data, str);
 	    }
 
-	    for (i = numITC * 2; i < bs->numTexop * 2; i++)
+	    for (i = j * 2; i < numIndirectOp * 2; i++)
 	    {
 		snprintf (str, 1024, "coord_%d", i);
 		ok &= addTempHeaderOpToFunctionData (data, str);
 	    }
 
+	    
 	    ok &= addFetchOpToFunctionData (data, "output", NULL, target);
 	    ok &= addColorOpToFunctionData (data, "output", "output");
 
-	    if (!numITC)
-	    {
-	        snprintf (str, 1024,
-			  "MUL fCoord, fragment.position, program.env[%d];",
-			  param);
+	    snprintf (str, 1024,
+		      "MUL fCoord, fragment.position, program.env[%d];",
+		      param);
 
-		ok &= addDataOpToFunctionData (data, str);
-	    }
-	    /* some drivers implement fragment.position as texture coordinate
-	       and we might already use them all */
-	    else if (numITC < bs->numTexop)
-	    {
-	        snprintf (str, 1024,
-			  "RCP fCoord, fragment.texcoord[%d].w;"
-			  "MUL fCoord, fCoord, fragment.texcoord[%d];",
-			  startTC, startTC);
+	    ok &= addDataOpToFunctionData (data, str);
 
-		ok &= addDataOpToFunctionData (data, str);
-	    }
+	    snprintf (str, 1024,
+		      "TEX sum, fCoord, texture[%d], %s;",
+		       unit + 1, targetString);
 
-	    if (numITC)
-	    {
-		snprintf (str, 1024,
-			  "TXP sum, fragment.texcoord[%d], texture[%d], %s;",
-			  startTC, unit + 1, targetString);
-
-		ok &= addDataOpToFunctionData (data, str);
-	    }
-	    else
-	    {
-		snprintf (str, 1024,
-			  "TEX sum, fCoord, texture[%d], %s;",
-			  unit + 1, targetString);
-
-		ok &= addDataOpToFunctionData (data, str);
-	    }
+	    ok &= addDataOpToFunctionData (data, str);
 
 	    snprintf (str, 1024,
 		      "MUL_SAT mask, output.a, program.env[%d];"
@@ -1250,51 +1243,60 @@ getDstBlurFragmentFunction (CompScreen  *s,
 
 	    ok &= addDataOpToFunctionData (data, str);
 
-	    for (i = numITC; i < bs->numTexop; i++)
+	    for (j = 0; j < numIndirect; j++)
 	    {
-		snprintf (str, 1024,
-			  "ADD coord_%d, fCoord, {0.0, %g, 0.0, 0.0};"
-			  "SUB coord_%d, fCoord, {0.0, %g, 0.0, 0.0};",
-			  i * 2, bs->pos[i] * bs->ty,
-			  (i * 2) + 1, bs->pos[i] * bs->ty);
+		base = j * numIndirectOp;
+		end  = MIN ((j + 1) * numIndirectOp, bs->numTexop) - base;
+		
+		ITCbase = MAX (numITC - base, 0);
 
-		ok &= addDataOpToFunctionData (data, str);
-	    }
-	    
-	    for (i = 0; i < numITC; i++)
-	    {
-		snprintf (str, 1024,
-			  "TXP pix_%d, fragment.texcoord[%d], texture[%d], %s;"
-			  "TXP pix_%d, fragment.texcoord[%d], texture[%d], %s;",
-			  i * 2, startTC + 1 + (i * 2),
-			  unit + 1, targetString,
-			  (i * 2) + 1, startTC + 2 + (i * 2),
-			  unit + 1, targetString);
+		for (i = ITCbase; i < end; i++)
+		{
+		    snprintf (str, 1024,
+			      "ADD coord_%d, fCoord, {0.0, %g, 0.0, 0.0};"
+			      "SUB coord_%d, fCoord, {0.0, %g, 0.0, 0.0};",
+			      i * 2, bs->pos[base + i] * bs->ty,
+			      (i * 2) + 1, bs->pos[base + i] * bs->ty);
 
-		ok &= addDataOpToFunctionData (data, str);
+		    ok &= addDataOpToFunctionData (data, str);
+		}
+
+		for (i = 0; i < ITCbase; i++)
+		{
+		    snprintf (str, 1024,
+			"TXP pix_%d, fragment.texcoord[%d], texture[%d], %s;"
+			"TXP pix_%d, fragment.texcoord[%d], texture[%d], %s;",
+			i * 2, startTC + ((i + base) * 2),
+			unit + 1, targetString,
+			(i * 2) + 1, startTC + 1 + ((i + base) * 2),
+			unit + 1, targetString);
+
+		    ok &= addDataOpToFunctionData (data, str);
+		}
+
+		for (i = ITCbase; i < end; i++)
+		{
+		    snprintf (str, 1024,
+			      "TEX pix_%d, coord_%d, texture[%d], %s;"
+			      "TEX pix_%d, coord_%d, texture[%d], %s;",
+			      i * 2, i * 2,
+			      unit + 1, targetString,
+			      (i * 2) + 1, (i * 2) + 1,
+			      unit + 1, targetString);
+
+		    ok &= addDataOpToFunctionData (data, str);
+		}
+
+		for (i = 0; i < end * 2; i++)
+		{
+		    snprintf (str, 1024,
+			      "MAD sum, pix_%d, %f, sum;",
+			      i, bs->amp[base + (i / 2)]);
+
+		    ok &= addDataOpToFunctionData (data, str);
+		}
 	    }
 
-	    for (i = numITC; i < bs->numTexop; i++)
-	    {
-		snprintf (str, 1024,
-			  "TEX pix_%d, coord_%d, texture[%d], %s;"
-			  "TEX pix_%d, coord_%d, texture[%d], %s;",
-			  i * 2, i * 2,
-			  unit + 1, targetString,
-			  (i * 2) + 1, (i * 2) + 1,
-			  unit + 1, targetString);
-
-		ok &= addDataOpToFunctionData (data, str);
-	    }
-
-	    for (i = 0; i < bs->numTexop * 2; i++)
-	    {
-		snprintf (str, 1024,
-			  "MAD sum, pix_%d, %f, sum;",
-			  i, bs->amp[i / 2]);
-
-		ok &= addDataOpToFunctionData (data, str);
-	    }
 	} break;
 	case BLUR_FILTER_MIPMAP:
 	    ok &= addFetchOpToFunctionData (data, "output", NULL, target);
@@ -1424,8 +1426,8 @@ loadFragmentProgram (CompScreen *s,
     glGetIntegerv (GL_PROGRAM_ERROR_POSITION_ARB, &errorPos);
     if (glGetError () != GL_NO_ERROR || errorPos != -1)
     {
-	compLogMessage (s->display, "blur", CompLogLevelError,
-			"Failed to load blur program");
+	compLogMessage ("blur", CompLogLevelError,
+			"Failed to load blur program %s", string);
 
 	(*s->deletePrograms) (1, program);
 	*program = 0;
@@ -1439,10 +1441,13 @@ loadFragmentProgram (CompScreen *s,
 static Bool
 loadFilterProgram (CompScreen *s, int numITC)
 {
-    char  buffer[2048];
+    char  buffer[4096];
     char  *targetString;
     char  *str = buffer;
-    int   i;
+    int   i, j;
+    int   numIndirect;
+    int   numIndirectOp;
+    int   base, end, ITCbase;
 
     BLUR_SCREEN (s);
 
@@ -1456,12 +1461,26 @@ loadFilterProgram (CompScreen *s, int numITC)
 		    "ATTRIB texcoord = fragment.texcoord[0];"
 		    "TEMP sum;");
 
-    str += sprintf (str, "TEMP tCoord, pix;");
+    if (bs->maxTemp - 1 > (bs->numTexop + (bs->numTexop - numITC)) * 2)
+    {
+	numIndirect   = 1;
+	numIndirectOp = bs->numTexop;
+    }
+    else
+    {
+	i = (bs->maxTemp - 1) / 4;
+	numIndirect = ceil ((float)bs->numTexop / (float)i);
+	numIndirectOp = ceil ((float)bs->numTexop / (float)numIndirect);
+    }
 
-    for (i = 0; i < bs->numTexop; i++)
+    /* we need to define all coordinate temporaries if we have
+       multiple indirection steps */
+    j = (numIndirect > 1) ? 0 : numITC;
+
+    for (i = 0; i < numIndirectOp; i++)
 	str += sprintf (str,"TEMP pix_%d, pix_%d;", i * 2, (i * 2) + 1);
 
-    for (i = numITC; i < bs->numTexop; i++)
+    for (i = j; i < numIndirectOp; i++)
 	str += sprintf (str,"TEMP coord_%d, coord_%d;", i * 2, (i * 2) + 1);
 
     str += sprintf (str,
@@ -1471,33 +1490,41 @@ loadFilterProgram (CompScreen *s, int numITC)
     str += sprintf (str,
 		    "MUL sum, sum, %f;",
 		    bs->amp[bs->numTexop]);
-    
-    for (i = numITC; i < bs->numTexop; i++)
-	str += sprintf (str,
-			"ADD coord_%d, texcoord, {%g, 0.0, 0.0, 0.0};"
-			"SUB coord_%d, texcoord, {%g, 0.0, 0.0, 0.0};",
-			i * 2, bs->pos[i] * bs->tx,
-			(i * 2) + 1, bs->pos[i] * bs->tx);
 
-    for (i = 0; i < numITC; i++)
-	str += sprintf (str,
-			"TEX pix_%d, fragment.texcoord[%d], texture[0], %s;"
-			"TEX pix_%d, fragment.texcoord[%d], texture[0], %s;",
-			i * 2, (i * 2) + 1, targetString,
-			(i * 2) + 1, (i * 2) + 2, targetString);
+    for (j = 0; j < numIndirect; j++)
+    {
+	base = j * numIndirectOp;
+	end  = MIN ((j + 1) * numIndirectOp, bs->numTexop) - base;
+	
+	ITCbase = MAX (numITC - base, 0);
 
-    for (i = numITC; i < bs->numTexop; i++)
-	str += sprintf (str,
-			"TEX pix_%d, coord_%d, texture[0], %s;"
-			"TEX pix_%d, coord_%d, texture[0], %s;",
-			i * 2, i * 2, targetString,
-			(i * 2) + 1, (i * 2) + 1, targetString);
+	for (i = ITCbase; i < end; i++)
+	    str += sprintf (str,
+			    "ADD coord_%d, texcoord, {%g, 0.0, 0.0, 0.0};"
+			    "SUB coord_%d, texcoord, {%g, 0.0, 0.0, 0.0};",
+			    i * 2, bs->pos[base + i] * bs->tx,
+			    (i * 2) + 1, bs->pos[base + i] * bs->tx);
+
+	for (i = 0; i < ITCbase; i++)
+	    str += sprintf (str,
+		"TEX pix_%d, fragment.texcoord[%d], texture[0], %s;"
+		"TEX pix_%d, fragment.texcoord[%d], texture[0], %s;",
+		i * 2, ((i + base) * 2) + 1, targetString,
+		(i * 2) + 1, ((i + base) * 2) + 2, targetString);
+
+	for (i = ITCbase; i < end; i++)
+	    str += sprintf (str,
+			    "TEX pix_%d, coord_%d, texture[0], %s;"
+			    "TEX pix_%d, coord_%d, texture[0], %s;",
+			    i * 2, i * 2, targetString,
+			    (i * 2) + 1, (i * 2) + 1, targetString);
+
+	for (i = 0; i < end * 2; i++)
+	    str += sprintf (str,
+			    "MAD sum, pix_%d, %f, sum;",
+			    i, bs->amp[base + (i / 2)]);
+    }
 
-    for (i = 0; i < bs->numTexop * 2; i++)
-	str += sprintf (str,
-			"MAD sum, pix_%d, %f, sum;",
-			i, bs->amp[i / 2]);
-    
     str += sprintf (str,
 		    "MOV result.color, sum;"
 		    "END");
@@ -1526,7 +1553,7 @@ fboPrologue (CompScreen *s)
 	bs->fboStatus = (*s->checkFramebufferStatus) (GL_FRAMEBUFFER_EXT);
 	if (bs->fboStatus != GL_FRAMEBUFFER_COMPLETE_EXT)
 	{
-	    compLogMessage (s->display, "blur", CompLogLevelError,
+	    compLogMessage ("blur", CompLogLevelError,
 			    "Framebuffer incomplete");
 
 	    (*s->bindFramebuffer) (GL_FRAMEBUFFER_EXT, 0);
@@ -1591,7 +1618,8 @@ fboUpdate (CompScreen *s,
 	   BoxPtr     pBox,
 	   int	      nBox)
 {
-    int i, y, iTC = 0;
+    int  i, y, iTC = 0;
+    Bool wasCulled = glIsEnabled (GL_CULL_FACE);
 
     BLUR_SCREEN (s);
 
@@ -1606,6 +1634,8 @@ fboUpdate (CompScreen *s,
     if (!fboPrologue (s))
 	return FALSE;
 
+    glDisable (GL_CULL_FACE);
+
     glDisableClientState (GL_TEXTURE_COORD_ARRAY);
 
     glBindTexture (bs->target, bs->texture[0]);
@@ -1682,6 +1712,9 @@ fboUpdate (CompScreen *s,
 
     glEnableClientState (GL_TEXTURE_COORD_ARRAY);
 
+    if (wasCulled)
+	glEnable (GL_CULL_FACE);
+
     fboEpilogue (s);
 
     return TRUE;
@@ -1697,7 +1730,7 @@ blurProjectRegion (CompWindow	       *w,
     CompScreen *s = w->screen;
     float      screen[MAX_VERTEX_PROJECT_COUNT * 2];
     float      vertices[MAX_VERTEX_PROJECT_COUNT * 3];
-    int	       nVertices;
+    int	       nVertices, nQuadCombine;
     int        i, j, stride;
     float      *v, *vert;
     float      minX, maxX, minY, maxY, minZ, maxZ;
@@ -1713,7 +1746,8 @@ blurProjectRegion (CompWindow	       *w,
     if (!w->vCount)
 	return;
 
-    nVertices = (w->indexCount) ? w->indexCount: w->vCount;
+    nVertices    = (w->indexCount) ? w->indexCount: w->vCount;
+    nQuadCombine = 1;
 
     stride = w->vertexStride;
     vert = w->vertices + (stride - 3);
@@ -1776,11 +1810,9 @@ blurProjectRegion (CompWindow	       *w,
 	vertices[2] = vertices[5]  = maxZ;
 	vertices[8] = vertices[11] = maxZ;
 
-	if (maxZ == minZ)
-	{
-	    nVertices = 4;
-	}
-	else
+	nVertices = 4;
+
+	if (maxZ != minZ)
 	{
 	    vertices[12] = vertices[21] = minX;
 	    vertices[13] = vertices[16] = minY;
@@ -1788,12 +1820,13 @@ blurProjectRegion (CompWindow	       *w,
 	    vertices[19] = vertices[22] = maxY;
 	    vertices[14] = vertices[17] = minZ;
 	    vertices[20] = vertices[23] = minZ;
-	    nVertices = 8;
+	    nQuadCombine = 2;
 	}
     }
 
+
     if (!projectVertices (w->screen, output, transform, vertices, screen,
-			  nVertices))
+			  nVertices * nQuadCombine))
 	return;
 
     region.rects    = &region.extents;
@@ -1808,7 +1841,7 @@ blurProjectRegion (CompWindow	       *w,
 	minY = s->height;
 	maxY = 0;
 
-	for (j = 0; j < 8; j += 2)
+	for (j = 0; j < 8 * nQuadCombine; j += 2)
 	{
 	    if (scr[j] < minX)
 		minX = scr[j];
@@ -1961,7 +1994,7 @@ blurUpdateDstTexture (CompWindow	  *w,
 		(*s->genFramebuffers) (1, &bs->fbo);
 
 	    if (!bs->fbo)
-		compLogMessage (s->display, "blur", CompLogLevelError,
+		compLogMessage ("blur", CompLogLevelError,
 				"Failed to create framebuffer object");
 
 	    textures = 2;
@@ -1996,15 +2029,15 @@ blurUpdateDstTexture (CompWindow	  *w,
 	    {
 		if (!s->fbo)
 		{
-		    compLogMessage (s->display, "blur", CompLogLevelWarn,
-			     "GL_EXT_framebuffer_object extension "
-			     "is required for mipmap filter");
+		    compLogMessage ("blur", CompLogLevelWarn,
+				    "GL_EXT_framebuffer_object extension "
+				    "is required for mipmap filter");
 		}
 		else if (bs->target != GL_TEXTURE_2D)
 		{
-		    compLogMessage (s->display, "blur", CompLogLevelWarn,
-			     "GL_ARB_texture_non_power_of_two "
-			     "extension is required for mipmap filter");
+		    compLogMessage ("blur", CompLogLevelWarn,
+				    "GL_ARB_texture_non_power_of_two "
+				    "extension is required for mipmap filter");
 		}
 		else
 		{
@@ -2287,9 +2320,10 @@ blurDrawWindowTexture (CompWindow	    *w,
 	    case BLUR_FILTER_GAUSSIAN:
 		if (bs->opt[BLUR_SCREEN_OPTION_INDEPENDENT_TEX].value.b)
 		{
-		    iTC = MAX (0, s->maxTextureUnits - w->texUnits);
+		    /* leave one free texture unit for fragment position */
+		    iTC = MAX (0, s->maxTextureUnits - (w->texUnits + 1));
 		    if (iTC)
-			iTC = MIN (((iTC - 1) / 2), bs->numTexop);
+			iTC = MIN (iTC / 2, bs->numTexop);
 		}
 
 		param = allocFragmentParameters (&dstFa, 2);
@@ -2340,37 +2374,10 @@ blurDrawWindowTexture (CompWindow	    *w,
 
 			matrixMultiply (&tm, &tm, &bs->mvp);
 
-			s_gen[0] = tm.m[0];
-			s_gen[1] = tm.m[4];
-			s_gen[2] = tm.m[8];
-			s_gen[3] = tm.m[12];
-			t_gen[0] = tm.m[1];
-			t_gen[1] = tm.m[5];
-			t_gen[2] = tm.m[9];
-			t_gen[3] = tm.m[13];
-			q_gen[0] = tm.m[3];
-			q_gen[1] = tm.m[7];
-			q_gen[2] = tm.m[11];
-			q_gen[3] = tm.m[15];
-
-			(*s->activeTexture) (GL_TEXTURE0_ARB + w->texUnits);
-
-			glTexGenfv(GL_T, GL_OBJECT_PLANE, t_gen);
-			glTexGenfv(GL_S, GL_OBJECT_PLANE, s_gen);
-			glTexGenfv(GL_Q, GL_OBJECT_PLANE, q_gen);
-
-			glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
-			glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
-			glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
-
-			glEnable(GL_TEXTURE_GEN_S);
-			glEnable(GL_TEXTURE_GEN_T);
-			glEnable(GL_TEXTURE_GEN_Q);
-
 			for (i = 0; i < iTC; i++)
 			{
 			    (*s->activeTexture) (GL_TEXTURE0_ARB + w->texUnits
-						 + 1 + (i * 2));
+						 + (i * 2));
 
 			    matrixGetIdentity (&rm);
 			    rm.m[13] = bs->ty * bs->pos[i];
@@ -2404,7 +2411,7 @@ blurDrawWindowTexture (CompWindow	    *w,
 			    glEnable(GL_TEXTURE_GEN_Q);
 
 			    (*s->activeTexture) (GL_TEXTURE0_ARB + w->texUnits
-						 + 2 + (i * 2));
+						 + 1 + (i * 2));
 
 			    matrixGetIdentity (&rm);
 			    rm.m[13] = -bs->ty * bs->pos[i];
@@ -2519,7 +2526,7 @@ blurDrawWindowTexture (CompWindow	    *w,
 	if (iTC)
 	{
 	    int i;
-	    for (i = w->texUnits; i < w->texUnits + 1 + (2 * iTC); i++)
+	    for (i = w->texUnits; i < w->texUnits + (2 * iTC); i++)
 	    {
 		(*s->activeTexture) (GL_TEXTURE0_ARB + i);
 		glDisable(GL_TEXTURE_GEN_S);
@@ -2981,12 +2988,13 @@ blurInitScreen (CompPlugin *p,
 	bs->texture[i] = 0;
 
     bs->program   = 0;
+    bs->maxTemp   = 32;
     bs->fbo	  = 0;
     bs->fboStatus = FALSE;
 
     glGetIntegerv (GL_STENCIL_BITS, &bs->stencilBits);
     if (!bs->stencilBits)
-	compLogMessage (s->display, "blur", CompLogLevelWarn,
+	compLogMessage ("blur", CompLogLevelWarn,
 			"No stencil buffer. Region based blur disabled");
 
     /* We need GL_ARB_fragment_program for blur */
@@ -2995,6 +3003,15 @@ blurInitScreen (CompPlugin *p,
     else
 	bs->alphaBlur = FALSE;
 
+    if (s->fragmentProgram)
+    {
+	int tmp[4];
+	s->getProgramiv (GL_FRAGMENT_PROGRAM_ARB,
+			 GL_MAX_PROGRAM_NATIVE_TEMPORARIES_ARB,
+			 tmp);
+	bs->maxTemp = tmp[0];
+    }
+
     WRAP (bs, s, preparePaintScreen, blurPreparePaintScreen);
     WRAP (bs, s, donePaintScreen, blurDonePaintScreen);
     WRAP (bs, s, paintOutput, blurPaintOutput);
