@@ -29,21 +29,9 @@
 
 #include <X11/cursorfont.h>
 
-#include <compiz.h>
+#include <compiz-core.h>
 
-#define MOVE_INITIATE_BUTTON_DEFAULT	       Button1
-#define MOVE_INITIATE_BUTTON_MODIFIERS_DEFAULT CompAltMask
-
-#define MOVE_INITIATE_KEY_DEFAULT	    "F7"
-#define MOVE_INITIATE_KEY_MODIFIERS_DEFAULT CompAltMask
-
-#define MOVE_OPACITY_DEFAULT 100
-#define MOVE_OPACITY_MIN     1
-#define MOVE_OPACITY_MAX     100
-
-#define MOVE_CONSTRAIN_Y_DEFAULT TRUE
-
-#define MOVE_SNAPOFF_MAXIMIZED_DEFAULT TRUE
+static CompMetadata moveMetadata;
 
 struct _MoveKeys {
     char *name;
@@ -65,11 +53,13 @@ struct _MoveKeys {
 
 static int displayPrivateIndex;
 
-#define MOVE_DISPLAY_OPTION_INITIATE	      0
-#define MOVE_DISPLAY_OPTION_OPACITY	      1
-#define MOVE_DISPLAY_OPTION_CONSTRAIN_Y	      2
-#define MOVE_DISPLAY_OPTION_SNAPOFF_MAXIMIZED 3
-#define MOVE_DISPLAY_OPTION_NUM		      4
+#define MOVE_DISPLAY_OPTION_INITIATE_BUTTON   0
+#define MOVE_DISPLAY_OPTION_INITIATE_KEY      1
+#define MOVE_DISPLAY_OPTION_OPACITY	      2
+#define MOVE_DISPLAY_OPTION_CONSTRAIN_Y	      3
+#define MOVE_DISPLAY_OPTION_SNAPOFF_MAXIMIZED 4
+#define MOVE_DISPLAY_OPTION_LAZY_POSITIONING  5
+#define MOVE_DISPLAY_OPTION_NUM		      6
 
 typedef struct _MoveDisplay {
     int		    screenPrivateIndex;
@@ -86,6 +76,8 @@ typedef struct _MoveDisplay {
     int        status;
     KeyCode    key[NUM_KEYS];
 
+    int releaseButton;
+
     GLushort moveOpacity;
 } MoveDisplay;
 
@@ -102,14 +94,14 @@ typedef struct _MoveScreen {
     int	snapBackY;
 } MoveScreen;
 
-#define GET_MOVE_DISPLAY(d)				     \
-    ((MoveDisplay *) (d)->privates[displayPrivateIndex].ptr)
+#define GET_MOVE_DISPLAY(d)					  \
+    ((MoveDisplay *) (d)->base.privates[displayPrivateIndex].ptr)
 
 #define MOVE_DISPLAY(d)		           \
     MoveDisplay *md = GET_MOVE_DISPLAY (d)
 
-#define GET_MOVE_SCREEN(s, md)				         \
-    ((MoveScreen *) (s)->privates[(md)->screenPrivateIndex].ptr)
+#define GET_MOVE_SCREEN(s, md)					      \
+    ((MoveScreen *) (s)->base.privates[(md)->screenPrivateIndex].ptr)
 
 #define MOVE_SCREEN(s)						        \
     MoveScreen *ms = GET_MOVE_SCREEN (s, GET_MOVE_DISPLAY (s->display))
@@ -131,11 +123,11 @@ moveInitiate (CompDisplay     *d,
     xid = getIntOptionNamed (option, nOption, "window", 0);
 
     w = findWindowAtDisplay (d, xid);
-    if (w)
+    if (w && (w->actions & CompWindowActionMoveMask))
     {
 	XRectangle   workArea;
 	unsigned int mods;
-	int          x, y;
+	int          x, y, button;
 
 	MOVE_SCREEN (w->screen);
 
@@ -146,6 +138,8 @@ moveInitiate (CompDisplay     *d,
 	y = getIntOptionNamed (option, nOption, "y",
 			       w->attrib.y + (w->height / 2));
 
+	button = getIntOptionNamed (option, nOption, "button", -1);
+
 	if (otherScreenGrabExist (w->screen, "move", 0))
 	    return FALSE;
 
@@ -196,6 +190,8 @@ moveInitiate (CompDisplay     *d,
 	{
 	    md->w = w;
 
+	    md->releaseButton = button;
+
 	    (w->screen->windowGrabNotify) (w, x, y, mods,
 					   CompWindowGrabMoveMask |
 					   CompWindowGrabButtonMask);
@@ -207,7 +203,7 @@ moveInitiate (CompDisplay     *d,
 		xRoot = w->attrib.x + (w->width  / 2);
 		yRoot = w->attrib.y + (w->height / 2);
 
-		warpPointer (d, xRoot - pointerX, yRoot - pointerY);
+		warpPointer (w->screen, xRoot - pointerX, yRoot - pointerY);
 	    }
 
 	    if (md->moveOpacity != OPAQUE)
@@ -231,8 +227,6 @@ moveTerminate (CompDisplay     *d,
     {
 	MOVE_SCREEN (md->w->screen);
 
-	(md->w->screen->windowUngrabNotify) (md->w);
-
 	if (state & CompActionStateCancel)
 	    moveWindow (md->w,
 			md->savedX - md->w->attrib.x,
@@ -241,6 +235,13 @@ moveTerminate (CompDisplay     *d,
 
 	syncWindowPosition (md->w);
 
+	/* update window attributes as window constraints may have
+	   changed - needed e.g. if a maximized window was moved
+	   to another output device */
+	updateWindowAttributes (md->w, CompStackingUpdateModeNone);
+
+	(md->w->screen->windowUngrabNotify) (md->w);
+
 	if (ms->grabIndex)
 	{
 	    removeScreenGrab (md->w->screen, ms->grabIndex, NULL);
@@ -250,7 +251,8 @@ moveTerminate (CompDisplay     *d,
 	if (md->moveOpacity != OPAQUE)
 	    addWindowDamage (md->w);
 
-	md->w = 0;
+	md->w             = 0;
+	md->releaseButton = 0;
     }
 
     action->state &= ~(CompActionStateTermKey | CompActionStateTermButton);
@@ -266,6 +268,9 @@ moveGetYConstrainRegion (CompScreen *s)
     CompWindow *w;
     Region     region;
     REGION     r;
+    XRectangle workArea;
+    BoxRec     extents;
+    int	       i;
 
     region = XCreateRegion ();
     if (!region)
@@ -276,33 +281,70 @@ moveGetYConstrainRegion (CompScreen *s)
 
     r.extents.x1 = MINSHORT;
     r.extents.y1 = 0;
-    r.extents.x2 = MAXSHORT;
+    r.extents.x2 = 0;
     r.extents.y2 = s->height;
 
     XUnionRegion (&r, region, region);
 
-    for (w = s->windows; w; w = w->next)
+    r.extents.x1 = s->width;
+    r.extents.x2 = MAXSHORT;
+
+    XUnionRegion (&r, region, region);
+
+    for (i = 0; i < s->nOutputDev; i++)
     {
-	if (!w->mapNum)
-	    continue;
+	XUnionRegion (&s->outputDev[i].region, region, region);
 
-	if (w->struts)
+	getWorkareaForOutput (s, i, &workArea);
+	extents = s->outputDev[i].region.extents;
+
+	for (w = s->windows; w; w = w->next)
 	{
-	    r.extents.x1 = w->struts->top.x;
-	    r.extents.y1 = w->struts->top.y;
-	    r.extents.x2 = r.extents.x1 + w->struts->top.width;
-	    r.extents.y2 = r.extents.y1 + w->struts->top.height;
+	    if (!w->mapNum)
+		continue;
 
-	    if (r.extents.y2 <= s->workArea.y)
-		XSubtractRegion (region, &r, region);
+	    if (w->struts)
+	    {
+		r.extents.x1 = w->struts->top.x;
+		r.extents.y1 = w->struts->top.y;
+		r.extents.x2 = r.extents.x1 + w->struts->top.width;
+		r.extents.y2 = r.extents.y1 + w->struts->top.height;
 
-	    r.extents.x1 = w->struts->bottom.x;
-	    r.extents.y1 = w->struts->bottom.y;
-	    r.extents.x2 = r.extents.x1 + w->struts->bottom.width;
-	    r.extents.y2 = r.extents.y1 + w->struts->bottom.height;
+		if (r.extents.x1 < extents.x1)
+		    r.extents.x1 = extents.x1;
+		if (r.extents.x2 > extents.x2)
+		    r.extents.x2 = extents.x2;
+		if (r.extents.y1 < extents.y1)
+		    r.extents.y1 = extents.y1;
+		if (r.extents.y2 > extents.y2)
+		    r.extents.y2 = extents.y2;
 
-	    if (r.extents.y1 >= (s->workArea.y + s->workArea.height))
-		XSubtractRegion (region, &r, region);
+		if (r.extents.x1 < r.extents.x2 && r.extents.y1 < r.extents.y2)
+		{
+		    if (r.extents.y2 <= workArea.y)
+			XSubtractRegion (region, &r, region);
+		}
+
+		r.extents.x1 = w->struts->bottom.x;
+		r.extents.y1 = w->struts->bottom.y;
+		r.extents.x2 = r.extents.x1 + w->struts->bottom.width;
+		r.extents.y2 = r.extents.y1 + w->struts->bottom.height;
+
+		if (r.extents.x1 < extents.x1)
+		    r.extents.x1 = extents.x1;
+		if (r.extents.x2 > extents.x2)
+		    r.extents.x2 = extents.x2;
+		if (r.extents.y1 < extents.y1)
+		    r.extents.y1 = extents.y1;
+		if (r.extents.y2 > extents.y2)
+		    r.extents.y2 = extents.y2;
+
+		if (r.extents.x1 < r.extents.x2 && r.extents.y1 < r.extents.y2)
+		{
+		    if (r.extents.y1 >= (workArea.y + workArea.height))
+			XSubtractRegion (region, &r, region);
+		}
+	    }
 	}
     }
 
@@ -320,11 +362,18 @@ moveHandleMotionEvent (CompScreen *s,
     {
 	CompWindow *w;
 	int	   dx, dy;
+	int	   wX, wY;
+	int	   wWidth, wHeight;
 
 	MOVE_DISPLAY (s->display);
 
 	w = md->w;
 
+	wX      = w->serverX;
+	wY      = w->serverY;
+	wWidth  = w->serverWidth  + w->serverBorderWidth * 2;
+	wHeight = w->serverHeight + w->serverBorderWidth * 2;
+
 	md->x += xRoot - lastPointerX;
 	md->y += yRoot - lastPointerY;
 
@@ -357,9 +406,9 @@ moveHandleMotionEvent (CompScreen *s,
 		    int x, y, width, height;
 		    int status;
 
-		    x	   = w->attrib.x + dx - w->input.left;
-		    y	   = w->attrib.y + dy - w->input.top;
-		    width  = w->width + w->input.left + w->input.right;
+		    x	   = wX + dx - w->input.left;
+		    y	   = wY + dy - w->input.top;
+		    width  = wWidth + w->input.left + w->input.right;
 		    height = w->input.top ? w->input.top : 1;
 
 		    status = XRectInRegion (md->region, x, y, width, height);
@@ -378,7 +427,7 @@ moveHandleMotionEvent (CompScreen *s,
 			    if (xStatus != RectangleIn)
 				dx += (dx < 0) ? 1 : -1;
 
-			    x = w->attrib.x + dx - w->input.left;
+			    x = wX + dx - w->input.left;
 			}
 
 			while (dy && status != RectangleIn)
@@ -390,7 +439,7 @@ moveHandleMotionEvent (CompScreen *s,
 			    if (status != RectangleIn)
 				dy += (dy < 0) ? 1 : -1;
 
-			    y = w->attrib.y + dy - w->input.top;
+			    y = wY + dy - w->input.top;
 			}
 		    }
 		    else
@@ -404,30 +453,33 @@ moveHandleMotionEvent (CompScreen *s,
 	    {
 		if (w->state & CompWindowStateMaximizedVertMask)
 		{
-		    if ((yRoot - workArea.y) - ms->snapOffY >= SNAP_OFF)
+		    if (abs ((yRoot - workArea.y) - ms->snapOffY) >= SNAP_OFF)
 		    {
-			int width = w->serverWidth;
+			if (!otherScreenGrabExist (s, "move", 0))
+			{
+			    int width = w->serverWidth;
 
-			w->saveMask |= CWX | CWY;
+			    w->saveMask |= CWX | CWY;
 
-			if (w->saveMask & CWWidth)
-			    width = w->saveWc.width;
+			    if (w->saveMask & CWWidth)
+				width = w->saveWc.width;
 
-			w->saveWc.x = xRoot - (width >> 1);
-			w->saveWc.y = yRoot + (w->input.top >> 1);
+			    w->saveWc.x = xRoot - (width >> 1);
+			    w->saveWc.y = yRoot + (w->input.top >> 1);
 
-			md->x = md->y = 0;
+			    md->x = md->y = 0;
 
-			maximizeWindow (w, 0);
+			    maximizeWindow (w, 0);
 
-			ms->snapOffY = ms->snapBackY;
+			    ms->snapOffY = ms->snapBackY;
 
-			return;
+			    return;
+			}
 		    }
 		}
 		else if (ms->origState & CompWindowStateMaximizedVertMask)
 		{
-		    if ((yRoot - workArea.y) - ms->snapBackY < SNAP_BACK)
+		    if (abs ((yRoot - workArea.y) - ms->snapBackY) < SNAP_BACK)
 		    {
 			if (!otherScreenGrabExist (s, "move", 0))
 			{
@@ -443,7 +495,7 @@ moveHandleMotionEvent (CompScreen *s,
 			    wy  = workArea.y + (w->input.top >> 1);
 			    wy += w->sizeHints.height_inc >> 1;
 
-			    warpPointer (s->display, 0, wy - pointerY);
+			    warpPointer (s, 0, wy - pointerY);
 
 			    return;
 			}
@@ -454,40 +506,55 @@ moveHandleMotionEvent (CompScreen *s,
 	    if (w->state & CompWindowStateMaximizedVertMask)
 	    {
 		min = workArea.y + w->input.top;
-		max = workArea.y + workArea.height -
-		    w->input.bottom - w->serverHeight -
-		    w->serverBorderWidth * 2;
+		max = workArea.y + workArea.height - w->input.bottom - wHeight;
 
-		if (w->attrib.y + dy < min)
-		    dy = min - w->attrib.y;
-		else if (w->attrib.y + dy > max)
-		    dy = max - w->attrib.y;
+		if (wY + dy < min)
+		    dy = min - wY;
+		else if (wY + dy > max)
+		    dy = max - wY;
 	    }
 
 	    if (w->state & CompWindowStateMaximizedHorzMask)
 	    {
-		if (w->attrib.x > s->width || w->attrib.x + w->width < 0)
+		if (wX > s->width || wX + w->width < 0)
 		    return;
 
-		if (w->attrib.x + w->serverWidth + w->serverBorderWidth < 0)
+		if (wX + wWidth < 0)
 		    return;
 
 		min = workArea.x + w->input.left;
-		max = workArea.x + workArea.width -
-		    w->input.right - w->serverWidth -
-		    w->serverBorderWidth * 2;
+		max = workArea.x + workArea.width - w->input.right - wWidth;
 
-		if (w->attrib.x + dx < min)
-		    dx = min - w->attrib.x;
-		else if (w->attrib.x + dx > max)
-		    dx = max - w->attrib.x;
+		if (wX + dx < min)
+		    dx = min - wX;
+		else if (wX + dx > max)
+		    dx = max - wX;
 	    }
 	}
 
-	moveWindow (md->w, dx, dy, TRUE, FALSE);
+	if (dx || dy)
+	{
+	    moveWindow (w,
+			wX + dx - w->attrib.x,
+			wY + dy - w->attrib.y,
+			TRUE, FALSE);
 
-	md->x -= dx;
-	md->y -= dy;
+	    if (md->opt[MOVE_DISPLAY_OPTION_LAZY_POSITIONING].value.b)
+	    {
+		/* FIXME: This form of lazy positioning is broken and should
+		   be replaced asap. Current code exists just to avoid a
+		   major performance regression in the 0.5.2 release. */
+		w->serverX = w->attrib.x;
+		w->serverY = w->attrib.y;
+	    }
+	    else
+	    {
+		syncWindowPosition (w);
+	    }
+
+	    md->x -= dx;
+	    md->y -= dy;
+	}
     }
 }
 
@@ -501,6 +568,7 @@ moveHandleEvent (CompDisplay *d,
 
     switch (event->type) {
     case ButtonPress:
+    case ButtonRelease:
 	s = findScreenAtDisplay (d, event->xbutton.root);
 	if (s)
 	{
@@ -508,20 +576,26 @@ moveHandleEvent (CompDisplay *d,
 
 	    if (ms->grabIndex)
 	    {
-		moveTerminate (d,
-			       &md->opt[MOVE_DISPLAY_OPTION_INITIATE].value.action,
-			       0, NULL, 0);
+		if (md->releaseButton == -1 ||
+		    md->releaseButton == event->xbutton.button)
+		{
+		    CompAction *action;
+		    int        opt = MOVE_DISPLAY_OPTION_INITIATE_BUTTON;
+
+		    action = &md->opt[opt].value.action;
+		    moveTerminate (d, action, CompActionStateTermButton,
+				   NULL, 0);
+		}
 	    }
 	}
 	break;
     case KeyPress:
-    case KeyRelease:
 	s = findScreenAtDisplay (d, event->xkey.root);
 	if (s)
 	{
 	    MOVE_SCREEN (s);
 
-	    if (ms->grabIndex && event->type == KeyPress)
+	    if (ms->grabIndex)
 	    {
 		int i;
 
@@ -560,10 +634,9 @@ moveHandleEvent (CompDisplay *d,
 		w = findWindowAtDisplay (d, event->xclient.window);
 		if (w)
 		{
-		    CompOption o[4];
+		    CompOption o[5];
 		    int	       xRoot, yRoot;
-		    CompAction *action =
-			&md->opt[MOVE_DISPLAY_OPTION_INITIATE].value.action;
+		    int	       option;
 
 		    o[0].type    = CompOptionTypeInt;
 		    o[0].name    = "window";
@@ -571,7 +644,9 @@ moveHandleEvent (CompDisplay *d,
 
 		    if (event->xclient.data.l[2] == WmMoveResizeMoveKeyboard)
 		    {
-			moveInitiate (d, action,
+			option = MOVE_DISPLAY_OPTION_INITIATE_KEY;
+
+			moveInitiate (d, &md->opt[option].value.action,
 				      CompActionStateInitKey,
 				      o, 1);
 		    }
@@ -581,6 +656,8 @@ moveHandleEvent (CompDisplay *d,
 			Window	     root, child;
 			int	     i;
 
+			option = MOVE_DISPLAY_OPTION_INITIATE_BUTTON;
+
 			XQueryPointer (d->display, w->screen->root,
 				       &root, &child, &xRoot, &yRoot,
 				       &i, &i, &mods);
@@ -600,29 +677,68 @@ moveHandleEvent (CompDisplay *d,
 			    o[3].name	 = "y";
 			    o[3].value.i = event->xclient.data.l[1];
 
+			    o[4].type    = CompOptionTypeInt;
+			    o[4].name    = "button";
+			    o[4].value.i = event->xclient.data.l[3] ?
+				           event->xclient.data.l[3] : -1;
+
 			    moveInitiate (d,
-					  action,
+					  &md->opt[option].value.action,
 					  CompActionStateInitButton,
-					  o, 4);
+					  o, 5);
 
 			    moveHandleMotionEvent (w->screen, xRoot, yRoot);
 			}
 		    }
 		}
 	    }
+	    else if (md->w && event->xclient.data.l[2] == WmMoveResizeCancel)
+	    {
+		if (md->w->id == event->xclient.window)
+		{
+		    int option;
+
+		    option = MOVE_DISPLAY_OPTION_INITIATE_BUTTON;
+		    moveTerminate (d,
+				   &md->opt[option].value.action,
+				   CompActionStateCancel, NULL, 0);
+		    option = MOVE_DISPLAY_OPTION_INITIATE_KEY;
+		    moveTerminate (d,
+				   &md->opt[option].value.action,
+				   CompActionStateCancel, NULL, 0);
+		}
+	    }
 	}
 	break;
     case DestroyNotify:
 	if (md->w && md->w->id == event->xdestroywindow.window)
+	{
+	    int option;
+
+	    option = MOVE_DISPLAY_OPTION_INITIATE_BUTTON;
 	    moveTerminate (d,
-			   &md->opt[MOVE_DISPLAY_OPTION_INITIATE].value.action,
+			   &md->opt[option].value.action,
 			   0, NULL, 0);
+	    option = MOVE_DISPLAY_OPTION_INITIATE_KEY;
+	    moveTerminate (d,
+			   &md->opt[option].value.action,
+			   0, NULL, 0);
+	}
 	break;
     case UnmapNotify:
 	if (md->w && md->w->id == event->xunmap.window)
+	{
+	    int option;
+
+	    option = MOVE_DISPLAY_OPTION_INITIATE_BUTTON;
 	    moveTerminate (d,
-			   &md->opt[MOVE_DISPLAY_OPTION_INITIATE].value.action,
+			   &md->opt[option].value.action,
 			   0, NULL, 0);
+	    option = MOVE_DISPLAY_OPTION_INITIATE_KEY;
+	    moveTerminate (d,
+			   &md->opt[option].value.action,
+			   0, NULL, 0);
+	}
     default:
 	break;
     }
@@ -666,58 +782,9 @@ movePaintWindow (CompWindow		 *w,
     return status;
 }
 
-static void
-moveDisplayInitOptions (MoveDisplay *md,
-			Display     *display)
-{
-    CompOption *o;
-
-    o = &md->opt[MOVE_DISPLAY_OPTION_INITIATE];
-    o->name			     = "initiate";
-    o->shortDesc		     = N_("Initiate Window Move");
-    o->longDesc			     = N_("Start moving window");
-    o->type			     = CompOptionTypeAction;
-    o->value.action.initiate	     = moveInitiate;
-    o->value.action.terminate	     = moveTerminate;
-    o->value.action.bell	     = FALSE;
-    o->value.action.edgeMask	     = 0;
-    o->value.action.type	     = CompBindingTypeButton;
-    o->value.action.state	     = CompActionStateInitButton;
-    o->value.action.button.modifiers = MOVE_INITIATE_BUTTON_MODIFIERS_DEFAULT;
-    o->value.action.button.button    = MOVE_INITIATE_BUTTON_DEFAULT;
-    o->value.action.type	    |= CompBindingTypeKey;
-    o->value.action.state	    |= CompActionStateInitKey;
-    o->value.action.key.modifiers    = MOVE_INITIATE_KEY_MODIFIERS_DEFAULT;
-    o->value.action.key.keycode      =
-	XKeysymToKeycode (display, XStringToKeysym (MOVE_INITIATE_KEY_DEFAULT));
-
-    o = &md->opt[MOVE_DISPLAY_OPTION_OPACITY];
-    o->name	  = "opacity";
-    o->shortDesc  = N_("Opacity");
-    o->longDesc	  = N_("Opacity level of moving windows");
-    o->type	  = CompOptionTypeInt;
-    o->value.i	  = MOVE_OPACITY_DEFAULT;
-    o->rest.i.min = MOVE_OPACITY_MIN;
-    o->rest.i.max = MOVE_OPACITY_MAX;
-
-    o = &md->opt[MOVE_DISPLAY_OPTION_CONSTRAIN_Y];
-    o->name	  = "constrain_y";
-    o->shortDesc  = N_("Constrain Y");
-    o->longDesc	  = N_("Constrain Y coordinate to workspace area");
-    o->type	  = CompOptionTypeBool;
-    o->value.b    = MOVE_CONSTRAIN_Y_DEFAULT;
-
-    o = &md->opt[MOVE_DISPLAY_OPTION_SNAPOFF_MAXIMIZED];
-    o->name      = "snapoff_maximized";
-    o->shortDesc = N_("Snapoff maximized windows");
-    o->longDesc  = N_("Snapoff and auto unmaximized maximized windows "
-		      "when dragging");
-    o->type      = CompOptionTypeBool;
-    o->value.b   = MOVE_SNAPOFF_MAXIMIZED_DEFAULT;
-}
-
 static CompOption *
-moveGetDisplayOptions (CompDisplay *display,
+moveGetDisplayOptions (CompPlugin  *plugin,
+		       CompDisplay *display,
 		       int	   *count)
 {
     MOVE_DISPLAY (display);
@@ -727,8 +794,9 @@ moveGetDisplayOptions (CompDisplay *display,
 }
 
 static Bool
-moveSetDisplayOption (CompDisplay    *display,
-		      char	     *name,
+moveSetDisplayOption (CompPlugin      *plugin,
+		      CompDisplay     *display,
+		      const char      *name,
 		      CompOptionValue *value)
 {
     CompOption *o;
@@ -741,10 +809,6 @@ moveSetDisplayOption (CompDisplay    *display,
 	return FALSE;
 
     switch (index) {
-    case MOVE_DISPLAY_OPTION_INITIATE:
-	if (setDisplayAction (display, o, value))
-	    return TRUE;
-	break;
     case MOVE_DISPLAY_OPTION_OPACITY:
 	if (compSetIntOption (o, value))
 	{
@@ -752,20 +816,22 @@ moveSetDisplayOption (CompDisplay    *display,
 	    return TRUE;
 	}
 	break;
-    case MOVE_DISPLAY_OPTION_CONSTRAIN_Y:
-	if (compSetBoolOption (o, value))
-	    return TRUE;
-	break;
-    case MOVE_DISPLAY_OPTION_SNAPOFF_MAXIMIZED:
-	if (compSetBoolOption (o, value))
-	    return TRUE;
     default:
-	break;
+	return compSetDisplayOption (display, o, value);
     }
 
     return FALSE;
 }
 
+static const CompMetadataOptionInfo moveDisplayOptionInfo[] = {
+    { "initiate_button", "button", 0, moveInitiate, moveTerminate },
+    { "initiate_key", "key", 0, moveInitiate, moveTerminate },
+    { "opacity", "int", "<min>0</min><max>100</max>", 0, 0 },
+    { "constrain_y", "bool", 0, 0, 0 },
+    { "snapoff_maximized", "bool", 0, 0, 0 },
+    { "lazy_positioning", "bool", 0, 0, 0 }
+};
+
 static Bool
 moveInitDisplay (CompPlugin  *p,
 		 CompDisplay *d)
@@ -773,24 +839,38 @@ moveInitDisplay (CompPlugin  *p,
     MoveDisplay *md;
     int	        i;
 
+    if (!checkPluginABI ("core", CORE_ABIVERSION))
+	return FALSE;
+
     md = malloc (sizeof (MoveDisplay));
     if (!md)
 	return FALSE;
 
-    md->screenPrivateIndex = allocateScreenPrivateIndex (d);
-    if (md->screenPrivateIndex < 0)
+    if (!compInitDisplayOptionsFromMetadata (d,
+					     &moveMetadata,
+					     moveDisplayOptionInfo,
+					     md->opt,
+					     MOVE_DISPLAY_OPTION_NUM))
     {
 	free (md);
 	return FALSE;
     }
 
-    md->moveOpacity = (MOVE_OPACITY_DEFAULT * OPAQUE) / 100;
+    md->screenPrivateIndex = allocateScreenPrivateIndex (d);
+    if (md->screenPrivateIndex < 0)
+    {
+	compFiniDisplayOptions (d, md->opt, MOVE_DISPLAY_OPTION_NUM);
+	free (md);
+	return FALSE;
+    }
 
-    moveDisplayInitOptions (md, d->display);
+    md->moveOpacity =
+	(md->opt[MOVE_DISPLAY_OPTION_OPACITY].value.i * OPAQUE) / 100;
 
-    md->w      = 0;
-    md->region = NULL;
-    md->status = RectangleOut;
+    md->w             = 0;
+    md->region        = NULL;
+    md->status        = RectangleOut;
+    md->releaseButton = 0;
 
     for (i = 0; i < NUM_KEYS; i++)
 	md->key[i] = XKeysymToKeycode (d->display,
@@ -798,7 +878,7 @@ moveInitDisplay (CompPlugin  *p,
 
     WRAP (md, d, handleEvent, moveHandleEvent);
 
-    d->privates[displayPrivateIndex].ptr = md;
+    d->base.privates[displayPrivateIndex].ptr = md;
 
     return TRUE;
 }
@@ -813,6 +893,8 @@ moveFiniDisplay (CompPlugin  *p,
 
     UNWRAP (md, d, handleEvent);
 
+    compFiniDisplayOptions (d, md->opt, MOVE_DISPLAY_OPTION_NUM);
+
     free (md);
 }
 
@@ -830,13 +912,11 @@ moveInitScreen (CompPlugin *p,
 
     ms->grabIndex = 0;
 
-    ms->moveCursor = XCreateFontCursor (s->display->display, XC_plus);
-
-    addScreenAction (s, &md->opt[MOVE_DISPLAY_OPTION_INITIATE].value.action);
+    ms->moveCursor = XCreateFontCursor (s->display->display, XC_fleur);
 
     WRAP (ms, s, paintWindow, movePaintWindow);
 
-    s->privates[md->screenPrivateIndex].ptr = ms;
+    s->base.privates[md->screenPrivateIndex].ptr = ms;
 
     return TRUE;
 }
@@ -849,15 +929,85 @@ moveFiniScreen (CompPlugin *p,
 
     UNWRAP (ms, s, paintWindow);
 
+    if (ms->moveCursor)
+	XFreeCursor (s->display->display, ms->moveCursor);
+
     free (ms);
 }
 
+static CompBool
+moveInitObject (CompPlugin *p,
+		CompObject *o)
+{
+    static InitPluginObjectProc dispTab[] = {
+	(InitPluginObjectProc) 0, /* InitCore */
+	(InitPluginObjectProc) moveInitDisplay,
+	(InitPluginObjectProc) moveInitScreen
+    };
+
+    RETURN_DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), TRUE, (p, o));
+}
+
+static void
+moveFiniObject (CompPlugin *p,
+		CompObject *o)
+{
+    static FiniPluginObjectProc dispTab[] = {
+	(FiniPluginObjectProc) 0, /* FiniCore */
+	(FiniPluginObjectProc) moveFiniDisplay,
+	(FiniPluginObjectProc) moveFiniScreen
+    };
+
+    DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), (p, o));
+}
+
+static CompOption *
+moveGetObjectOptions (CompPlugin *plugin,
+		      CompObject *object,
+		      int	 *count)
+{
+    static GetPluginObjectOptionsProc dispTab[] = {
+	(GetPluginObjectOptionsProc) 0, /* GetCoreOptions */
+	(GetPluginObjectOptionsProc) moveGetDisplayOptions
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab),
+		     (void *) (*count = 0), (plugin, object, count));
+}
+
+static CompBool
+moveSetObjectOption (CompPlugin      *plugin,
+		     CompObject      *object,
+		     const char      *name,
+		     CompOptionValue *value)
+{
+    static SetPluginObjectOptionProc dispTab[] = {
+	(SetPluginObjectOptionProc) 0, /* SetCoreOption */
+	(SetPluginObjectOptionProc) moveSetDisplayOption
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab), FALSE,
+		     (plugin, object, name, value));
+}
+
 static Bool
 moveInit (CompPlugin *p)
 {
+    if (!compInitPluginMetadataFromInfo (&moveMetadata,
+					 p->vTable->name,
+					 moveDisplayOptionInfo,
+					 MOVE_DISPLAY_OPTION_NUM,
+					 0, 0))
+	return FALSE;
+
     displayPrivateIndex = allocateDisplayPrivateIndex ();
     if (displayPrivateIndex < 0)
+    {
+	compFiniMetadata (&moveMetadata);
 	return FALSE;
+    }
+
+    compAddMetadataFromFile (&moveMetadata, p->vTable->name);
 
     return TRUE;
 }
@@ -865,42 +1015,29 @@ moveInit (CompPlugin *p)
 static void
 moveFini (CompPlugin *p)
 {
-    if (displayPrivateIndex >= 0)
-	freeDisplayPrivateIndex (displayPrivateIndex);
+    freeDisplayPrivateIndex (displayPrivateIndex);
+    compFiniMetadata (&moveMetadata);
 }
 
-static int
-moveGetVersion (CompPlugin *plugin,
-		int	   version)
+static CompMetadata *
+moveGetMetadata (CompPlugin *plugin)
 {
-    return ABIVERSION;
+    return &moveMetadata;
 }
 
 CompPluginVTable moveVTable = {
     "move",
-    N_("Move Window"),
-    N_("Move window"),
-    moveGetVersion,
+    moveGetMetadata,
     moveInit,
     moveFini,
-    moveInitDisplay,
-    moveFiniDisplay,
-    moveInitScreen,
-    moveFiniScreen,
-    0, /* InitWindow */
-    0, /* FiniWindow */
-    moveGetDisplayOptions,
-    moveSetDisplayOption,
-    0, /* GetScreenOptions */
-    0, /* SetScreenOption */
-    0, /* Deps */
-    0, /* nDeps */
-    0, /* Features */
-    0  /* nFeatures */
+    moveInitObject,
+    moveFiniObject,
+    moveGetObjectOptions,
+    moveSetObjectOption
 };
 
 CompPluginVTable *
-getCompPluginInfo (void)
+getCompPluginInfo20070830 (void)
 {
     return &moveVTable;
 }
