@@ -31,7 +31,7 @@
 #include <string.h>
 #include <math.h>
 
-#include <compiz.h>
+#include <compiz-core.h>
 
 #define WIN_X(w) ((w)->attrib.x - (w)->output.left)
 #define WIN_Y(w) ((w)->attrib.y - (w)->output.top)
@@ -104,64 +104,15 @@ typedef struct _Model {
     unsigned int snapCnt[4];
 } Model;
 
-#define WOBBLY_FRICTION_DEFAULT    3.0f
-#define WOBBLY_FRICTION_MIN        0.1f
-#define WOBBLY_FRICTION_MAX       10.0f
-#define WOBBLY_FRICTION_PRECISION  0.1f
+#define WOBBLY_EFFECT_NONE   0
+#define WOBBLY_EFFECT_SHIVER 1
+#define WOBBLY_EFFECT_LAST   WOBBLY_EFFECT_SHIVER
 
-#define WOBBLY_SPRING_K_DEFAULT    8.0f
-#define WOBBLY_SPRING_K_MIN        0.1f
-#define WOBBLY_SPRING_K_MAX       10.0f
-#define WOBBLY_SPRING_K_PRECISION  0.1f
-
-#define WOBBLY_GRID_RESOLUTION_DEFAULT  8
-#define WOBBLY_GRID_RESOLUTION_MIN      1
-#define WOBBLY_GRID_RESOLUTION_MAX      64
-
-#define WOBBLY_MIN_GRID_SIZE_DEFAULT  8
-#define WOBBLY_MIN_GRID_SIZE_MIN      4
-#define WOBBLY_MIN_GRID_SIZE_MAX      128
-
-typedef enum {
-    WobblyEffectNone = 0,
-    WobblyEffectShiver
-} WobblyEffect;
-
-static char *effectName[] = {
-    N_("None"),
-    N_("Shiver")
-};
-
-static WobblyEffect effectType[] = {
-    WobblyEffectNone,
-    WobblyEffectShiver
-};
-
-#define NUM_EFFECT (sizeof (effectType) / sizeof (effectType[0]))
-
-#define WOBBLY_MAP_DEFAULT   (effectName[0])
-#define WOBBLY_FOCUS_DEFAULT (effectName[0])
-
-#define WOBBLY_MAP_WINDOW_MATCH_DEFAULT \
-    "Splash | DropdownMenu | PopupMenu | Tooltip | Notification | Combo | " \
-    "Dnd | Unknown"
-
-#define WOBBLY_FOCUS_WINDOW_MATCH_DEFAULT ""
-
-#define WOBBLY_MOVE_WINDOW_MATCH_DEFAULT \
-    "Toolbar | Menu | Utility | Dialog | Normal | Unknown"
-
-#define WOBBLY_GRAB_WINDOW_MATCH_DEFAULT ""
-
-#define WOBBLY_SNAP_MODIFIERS_DEFAULT ShiftMask
-
-#define WOBBLY_SNAP_INVERTED_DEFAULT FALSE
-
-#define WOBBLY_MAXIMIZE_EFFECT_DEFAULT TRUE
+static CompMetadata wobblyMetadata;
 
 static int displayPrivateIndex;
 
-#define WOBBLY_DISPLAY_OPTION_SNAP          0
+#define WOBBLY_DISPLAY_OPTION_SNAP_KEY      0
 #define WOBBLY_DISPLAY_OPTION_SNAP_INVERTED 1
 #define WOBBLY_DISPLAY_OPTION_SHIVER        2
 #define WOBBLY_DISPLAY_OPTION_NUM           3
@@ -193,7 +144,7 @@ typedef struct _WobblyScreen {
 
     PreparePaintScreenProc preparePaintScreen;
     DonePaintScreenProc	   donePaintScreen;
-    PaintScreenProc	   paintScreen;
+    PaintOutputProc	   paintOutput;
     PaintWindowProc	   paintWindow;
     DamageWindowRectProc   damageWindowRect;
     AddWindowGeometryProc  addWindowGeometry;
@@ -207,9 +158,6 @@ typedef struct _WobblyScreen {
 
     Bool wobblyWindows;
 
-    WobblyEffect mapEffect;
-    WobblyEffect focusEffect;
-
     unsigned int grabMask;
     CompWindow	 *grabWindow;
     Bool         moveWindow;
@@ -227,20 +175,20 @@ typedef struct _WobblyWindow {
     unsigned int state;
 } WobblyWindow;
 
-#define GET_WOBBLY_DISPLAY(d)				       \
-    ((WobblyDisplay *) (d)->privates[displayPrivateIndex].ptr)
+#define GET_WOBBLY_DISPLAY(d)					    \
+    ((WobblyDisplay *) (d)->base.privates[displayPrivateIndex].ptr)
 
 #define WOBBLY_DISPLAY(d)		       \
     WobblyDisplay *wd = GET_WOBBLY_DISPLAY (d)
 
-#define GET_WOBBLY_SCREEN(s, wd)				   \
-    ((WobblyScreen *) (s)->privates[(wd)->screenPrivateIndex].ptr)
+#define GET_WOBBLY_SCREEN(s, wd)					\
+    ((WobblyScreen *) (s)->base.privates[(wd)->screenPrivateIndex].ptr)
 
 #define WOBBLY_SCREEN(s)						      \
     WobblyScreen *ws = GET_WOBBLY_SCREEN (s, GET_WOBBLY_DISPLAY (s->display))
 
-#define GET_WOBBLY_WINDOW(w, ws)				   \
-    ((WobblyWindow *) (w)->privates[(ws)->windowPrivateIndex].ptr)
+#define GET_WOBBLY_WINDOW(w, ws)					\
+    ((WobblyWindow *) (w)->base.privates[(ws)->windowPrivateIndex].ptr)
 
 #define WOBBLY_WINDOW(w)				         \
     WobblyWindow *ww = GET_WOBBLY_WINDOW  (w,		         \
@@ -250,7 +198,8 @@ typedef struct _WobblyWindow {
 #define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))
 
 static CompOption *
-wobblyGetScreenOptions (CompScreen *screen,
+wobblyGetScreenOptions (CompPlugin *plugin,
+			CompScreen *screen,
 			int	   *count)
 {
     WOBBLY_SCREEN (screen);
@@ -260,186 +209,35 @@ wobblyGetScreenOptions (CompScreen *screen,
 }
 
 static Bool
-wobblySetScreenOption (CompScreen      *screen,
-		     char	     *name,
-		     CompOptionValue *value)
+wobblySetScreenOption (CompPlugin      *plugin,
+		       CompScreen      *screen,
+		       const char      *name,
+		       CompOptionValue *value)
 {
     CompOption *o;
-    int	       index;
 
     WOBBLY_SCREEN (screen);
 
-    o = compFindOption (ws->opt, NUM_OPTIONS (ws), name, &index);
+    o = compFindOption (ws->opt, NUM_OPTIONS (ws), name, NULL);
     if (!o)
 	return FALSE;
 
-    switch (index) {
-    case WOBBLY_SCREEN_OPTION_FRICTION:
-    case WOBBLY_SCREEN_OPTION_SPRING_K:
-	if (compSetFloatOption (o, value))
-	    return TRUE;
-	break;
-    case WOBBLY_SCREEN_OPTION_GRID_RESOLUTION:
-	if (compSetIntOption (o, value))
-	    return TRUE;
-	break;
-    case WOBBLY_SCREEN_OPTION_MIN_GRID_SIZE:
-	if (compSetIntOption (o, value))
-	    return TRUE;
-	break;
-    case WOBBLY_SCREEN_OPTION_MAP_EFFECT:
-	if (compSetStringOption (o, value))
-	{
-	    int i;
-
-	    for (i = 0; i < NUM_EFFECT; i++)
-	    {
-		if (strcmp (o->value.s, effectName[i]) == 0)
-		{
-		    ws->mapEffect = effectType[i];
-		    return TRUE;
-		}
-	    }
-	}
-	break;
-    case WOBBLY_SCREEN_OPTION_FOCUS_EFFECT:
-	if (compSetStringOption (o, value))
-	{
-	    int i;
-
-	    for (i = 0; i < NUM_EFFECT; i++)
-	    {
-		if (strcmp (o->value.s, effectName[i]) == 0)
-		{
-		    ws->focusEffect = effectType[i];
-		    return TRUE;
-		}
-	    }
-	}
-	break;
-    case WOBBLY_SCREEN_OPTION_MAP_WINDOW_MATCH:
-    case WOBBLY_SCREEN_OPTION_FOCUS_WINDOW_MATCH:
-    case WOBBLY_SCREEN_OPTION_MOVE_WINDOW_MATCH:
-    case WOBBLY_SCREEN_OPTION_GRAB_WINDOW_MATCH:
-	if (compSetMatchOption (o, value))
-	    return TRUE;
-	break;
-    case WOBBLY_SCREEN_OPTION_MAXIMIZE_EFFECT:
-	if (compSetBoolOption (o, value))
-	    return TRUE;
-    default:
-	break;
-    }
-
-    return FALSE;
+    return compSetScreenOption (screen, o, value);
 }
 
-static void
-wobblyScreenInitOptions (WobblyScreen *ws,
-			 Display      *display)
-{
-    CompOption *o;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_FRICTION];
-    o->name		= "friction";
-    o->shortDesc	= N_("Friction");
-    o->longDesc		= N_("Spring Friction");
-    o->type		= CompOptionTypeFloat;
-    o->value.f		= WOBBLY_FRICTION_DEFAULT;
-    o->rest.f.min	= WOBBLY_FRICTION_MIN;
-    o->rest.f.max	= WOBBLY_FRICTION_MAX;
-    o->rest.f.precision = WOBBLY_FRICTION_PRECISION;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_SPRING_K];
-    o->name		= "spring_k";
-    o->shortDesc	= N_("Spring K");
-    o->longDesc		= N_("Spring Konstant");
-    o->type		= CompOptionTypeFloat;
-    o->value.f		= WOBBLY_SPRING_K_DEFAULT;
-    o->rest.f.min	= WOBBLY_SPRING_K_MIN;
-    o->rest.f.max	= WOBBLY_SPRING_K_MAX;
-    o->rest.f.precision = WOBBLY_SPRING_K_PRECISION;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_GRID_RESOLUTION];
-    o->name	  = "grid_resolution";
-    o->shortDesc  = N_("Grid Resolution");
-    o->longDesc	  = N_("Vertex Grid Resolution");
-    o->type	  = CompOptionTypeInt;
-    o->value.i	  = WOBBLY_GRID_RESOLUTION_DEFAULT;
-    o->rest.i.min = WOBBLY_GRID_RESOLUTION_MIN;
-    o->rest.i.max = WOBBLY_GRID_RESOLUTION_MAX;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_MIN_GRID_SIZE];
-    o->name	  = "min_grid_size";
-    o->shortDesc  = N_("Minimum Grid Size");
-    o->longDesc	  = N_("Minimum Vertex Grid Size");
-    o->type	  = CompOptionTypeInt;
-    o->value.i	  = WOBBLY_MIN_GRID_SIZE_DEFAULT;
-    o->rest.i.min = WOBBLY_MIN_GRID_SIZE_MIN;
-    o->rest.i.max = WOBBLY_MIN_GRID_SIZE_MAX;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_MAP_EFFECT];
-    o->name	      = "map_effect";
-    o->shortDesc      = N_("Map Effect");
-    o->longDesc	      = N_("Map Window Effect");
-    o->type	      = CompOptionTypeString;
-    o->value.s	      = strdup (WOBBLY_MAP_DEFAULT);
-    o->rest.s.string  = effectName;
-    o->rest.s.nString = NUM_EFFECT;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_FOCUS_EFFECT];
-    o->name	      = "focus_effect";
-    o->shortDesc      = N_("Focus Effect");
-    o->longDesc	      = N_("Focus Window Effect");
-    o->type	      = CompOptionTypeString;
-    o->value.s	      = strdup (WOBBLY_FOCUS_DEFAULT);
-    o->rest.s.string  = effectName;
-    o->rest.s.nString = NUM_EFFECT;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_MAP_WINDOW_MATCH];
-    o->name	 = "map_window_match";
-    o->shortDesc = N_("Map Windows");
-    o->longDesc	 = N_("Windows that should wobble when mapped");
-    o->type	 = CompOptionTypeMatch;
-
-    matchInit (&o->value.match);
-    matchAddFromString (&o->value.match, WOBBLY_MAP_WINDOW_MATCH_DEFAULT);
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_FOCUS_WINDOW_MATCH];
-    o->name	 = "focus_window_match";
-    o->shortDesc = N_("Focus Windows");
-    o->longDesc	 = N_("Windows that should wobble when focused");
-    o->type	 = CompOptionTypeMatch;
-
-    matchInit (&o->value.match);
-    matchAddFromString (&o->value.match, WOBBLY_FOCUS_WINDOW_MATCH_DEFAULT);
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_MOVE_WINDOW_MATCH];
-    o->name	 = "move_window_match";
-    o->shortDesc = N_("Move Windows");
-    o->longDesc	 = N_("Windows that should wobble when moved");
-    o->type	 = CompOptionTypeMatch;
-
-    matchInit (&o->value.match);
-    matchAddFromString (&o->value.match, WOBBLY_MOVE_WINDOW_MATCH_DEFAULT);
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_GRAB_WINDOW_MATCH];
-    o->name	 = "grab_window_match";
-    o->shortDesc = N_("Grab Windows");
-    o->longDesc	 = N_("Windows that should wobble when grabbed");
-    o->type	 = CompOptionTypeMatch;
-
-    matchInit (&o->value.match);
-    matchAddFromString (&o->value.match, WOBBLY_GRAB_WINDOW_MATCH_DEFAULT);
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_MAXIMIZE_EFFECT];
-    o->name	  = "maximize_effect";
-    o->shortDesc  = N_("Maximize Effect");
-    o->longDesc	  = N_("Wobble effect when maximizing and unmaximizing "
-		       "windows");
-    o->type	  = CompOptionTypeBool;
-    o->value.b    = WOBBLY_MAXIMIZE_EFFECT_DEFAULT;
-}
+static const CompMetadataOptionInfo wobblyScreenOptionInfo[] = {
+    { "friction", "float", "<min>0.1</min>", 0, 0 },
+    { "spring_k", "float", "<min>0.1</min>", 0, 0 },
+    { "grid_resolution", "int", "<min>1</min><max>64</max>", 0, 0 },
+    { "min_grid_size", "int", "<min>4</min><max>128</max>", 0, 0 },
+    { "map_effect", "int", RESTOSTRING (0, WOBBLY_EFFECT_LAST), 0, 0 },
+    { "focus_effect", "int", RESTOSTRING (0, WOBBLY_EFFECT_LAST), 0, 0 },
+    { "map_window_match", "match", 0, 0, 0 },
+    { "focus_window_match", "match", 0, 0, 0 },
+    { "grab_window_match", "match", 0, 0, 0 },
+    { "move_window_match", "match", 0, 0, 0 },
+    { "maximize_effect", "bool", 0, 0, 0 }
+};
 
 #define SNAP_WINDOW_TYPE (CompWindowTypeNormalMask  | \
 			  CompWindowTypeToolbarMask | \
@@ -1337,7 +1135,10 @@ createModel (int	  x,
     model->numObjects = GRID_WIDTH * GRID_HEIGHT;
     model->objects = malloc (sizeof (Object) * model->numObjects);
     if (!model->objects)
+    {
+	free (model);
 	return 0;
+    }
 
     model->anchorObject = 0;
     model->numSprings = 0;
@@ -1518,7 +1319,7 @@ modelStepObject (CompWindow *window,
 		    object->position.y > object->vertEdge.end)
 		    findNextWestEdge (window, object);
 
-		if (object->vertEdge.snapped == FALSE ||
+		if (!object->vertEdge.snapped ||
 		    objectReleaseWestEdge (window, model, object))
 		{
 		    object->position.x += object->velocity.x;
@@ -1553,7 +1354,7 @@ modelStepObject (CompWindow *window,
 		    object->position.y > object->vertEdge.end)
 		    findNextEastEdge (window, object);
 
-		if (object->vertEdge.snapped == FALSE ||
+		if (!object->vertEdge.snapped ||
 		    objectReleaseEastEdge (window, model, object))
 		{
 		    object->position.x += object->velocity.x;
@@ -1591,7 +1392,7 @@ modelStepObject (CompWindow *window,
 		    object->position.x > object->horzEdge.end)
 		    findNextNorthEdge (window, object);
 
-		if (object->horzEdge.snapped == FALSE ||
+		if (!object->horzEdge.snapped ||
 		    objectReleaseNorthEdge (window, model, object))
 		{
 		    object->position.y += object->velocity.y;
@@ -1626,7 +1427,7 @@ modelStepObject (CompWindow *window,
 		    object->position.x > object->horzEdge.end)
 		    findNextSouthEdge (window, object);
 
-		if (object->horzEdge.snapped == FALSE ||
+		if (!object->horzEdge.snapped ||
 		    objectReleaseSouthEdge (window, model, object))
 		{
 		    object->position.y += object->velocity.y;
@@ -1963,12 +1764,12 @@ wobblyDrawWindowGeometry (CompWindow *w)
 {
     int     texUnit = w->texUnits;
     int     currentTexUnit = 0;
-    int     stride = (1 + texUnit) * 2;
-    GLfloat *vertices = w->vertices + (stride - 2);
+    int     stride = w->vertexStride;
+    GLfloat *vertices = w->vertices + (stride - 3);
 
     stride *= sizeof (GLfloat);
 
-    glVertexPointer (2, GL_FLOAT, stride, vertices);
+    glVertexPointer (3, GL_FLOAT, stride, vertices);
 
     while (texUnit--)
     {
@@ -1978,8 +1779,8 @@ wobblyDrawWindowGeometry (CompWindow *w)
 	    glEnableClientState (GL_TEXTURE_COORD_ARRAY);
 	    currentTexUnit = texUnit;
 	}
-	vertices -= 2;
-	glTexCoordPointer (2, GL_FLOAT, stride, vertices);
+	vertices -= w->texCoordSize;
+	glTexCoordPointer (w->texCoordSize, GL_FLOAT, stride, vertices);
     }
 
     glDrawElements (GL_QUADS, w->indexCount, GL_UNSIGNED_SHORT, w->indices);
@@ -2049,7 +1850,7 @@ wobblyAddWindowGeometry (CompWindow *w,
 
 	w->texUnits = nMatrix;
 
-	vSize = 2 + nMatrix * 2;
+	vSize = 3 + nMatrix * 2;
 
 	nVertices = w->vCount;
 	nIndices  = w->indexCount;
@@ -2134,6 +1935,7 @@ wobblyAddWindowGeometry (CompWindow *w,
 
 		    *v++ = deformedX;
 		    *v++ = deformedY;
+		    *v++ = 0.0;
 
 		    nVertices++;
 
@@ -2149,6 +1951,8 @@ wobblyAddWindowGeometry (CompWindow *w,
 	}
 
 	w->vCount	      = nVertices;
+	w->vertexStride       = vSize;
+	w->texCoordSize       = 2;
 	w->indexCount	      = nIndices;
 	w->drawWindowGeometry = wobblyDrawWindowGeometry;
     }
@@ -2324,7 +2128,8 @@ wobblyHandleEvent (CompDisplay *d,
 		Bool		    inverted;
 		unsigned int	    mods = 0xffffffff;
 
-		action   = &wd->opt[WOBBLY_DISPLAY_OPTION_SNAP].value.action;
+		action   =
+		    &wd->opt[WOBBLY_DISPLAY_OPTION_SNAP_KEY].value.action;
 		inverted = wd->opt[WOBBLY_DISPLAY_OPTION_SNAP_INVERTED].value.b;
 
 		if (action->type & CompBindingTypeKey)
@@ -2368,8 +2173,6 @@ wobblyHandleEvent (CompDisplay *d,
 
 		if (ww->state & MAXIMIZE_STATE)
 		{
-		    WOBBLY_WINDOW (ws->grabWindow);
-
 		    if (ww->model && ww->grabbed)
 		    {
 			int dx, dy;
@@ -2405,18 +2208,20 @@ wobblyHandleEvent (CompDisplay *d,
 	if (w && isWobblyWin (w))
 	{
 	    int mIndex;
+	    int focusEffect;
 
 	    WOBBLY_WINDOW (w);
 	    WOBBLY_SCREEN (w->screen);
 
-	    mIndex = WOBBLY_SCREEN_OPTION_FOCUS_WINDOW_MATCH;
+	    mIndex      = WOBBLY_SCREEN_OPTION_FOCUS_WINDOW_MATCH;
+	    focusEffect = ws->opt[WOBBLY_SCREEN_OPTION_FOCUS_EFFECT].value.i;
 
-	    if (ws->focusEffect				    &&
+	    if ((focusEffect != WOBBLY_EFFECT_NONE)         &&
 		matchEval (&ws->opt[mIndex].value.match, w) &&
 		wobblyEnsureModel (w))
 	    {
-		switch (ws->focusEffect) {
-		case WobblyEffectShiver:
+		switch (focusEffect) {
+		case WOBBLY_EFFECT_SHIVER:
 		    modelAdjustObjectsForShiver (ww->model,
 						 WIN_X (w),
 						 WIN_Y (w),
@@ -2475,21 +2280,23 @@ wobblyDamageWindowRect (CompWindow *w,
 	if (isWobblyWin (w))
 	{
 	    int mIndex;
+	    int mapEffect;
 
 	    WOBBLY_WINDOW (w);
 	    WOBBLY_SCREEN (w->screen);
 
-	    mIndex = WOBBLY_SCREEN_OPTION_MAP_WINDOW_MATCH;
+	    mIndex    = WOBBLY_SCREEN_OPTION_MAP_WINDOW_MATCH;
+	    mapEffect = ws->opt[WOBBLY_SCREEN_OPTION_MAP_EFFECT].value.i;
 
 	    if (ws->opt[WOBBLY_SCREEN_OPTION_MAXIMIZE_EFFECT].value.b)
 		wobblyEnsureModel (w);
 
-	    if (ws->mapEffect				    &&
+	    if ((mapEffect != WOBBLY_EFFECT_NONE)           &&
 		matchEval (&ws->opt[mIndex].value.match, w) &&
 		wobblyEnsureModel (w))
 	    {
-		switch (ws->mapEffect) {
-		case WobblyEffectShiver:
+		switch (mapEffect) {
+		case WOBBLY_EFFECT_SHIVER:
 		    modelAdjustObjectsForShiver (ww->model,
 						 WIN_X (w), WIN_Y (w),
 						 WIN_W (w), WIN_H (w));
@@ -2520,6 +2327,10 @@ wobblyWindowResizeNotify (CompWindow *w,
 
     if (ws->opt[WOBBLY_SCREEN_OPTION_MAXIMIZE_EFFECT].value.b &&
 	isWobblyWin (w)					      &&
+	/* prevent wobbling when shading maximized windows - assuming that
+	   the height difference shaded - non-shaded will hardly be -1 and
+	   a lack of wobbly animation in that corner case is tolerable */
+	(dheight != -1)                                       &&
 	((w->state | ww->state) & MAXIMIZE_STATE))
     {
 	ww->state &= ~MAXIMIZE_STATE;
@@ -2656,8 +2467,11 @@ wobblyWindowGrabNotify (CompWindow   *w,
 
     mIndex = WOBBLY_SCREEN_OPTION_MOVE_WINDOW_MATCH;
 
-    ws->grabMask   = mask;
-    ws->grabWindow = w;
+    if (!ws->grabWindow)
+    {
+	ws->grabMask   = mask;
+	ws->grabWindow = w;
+    }
     ws->moveWindow = FALSE;
 
     if ((mask & CompWindowGrabButtonMask)	    &&
@@ -2750,8 +2564,11 @@ wobblyWindowUngrabNotify (CompWindow *w)
     WOBBLY_SCREEN (w->screen);
     WOBBLY_WINDOW (w);
 
-    ws->grabMask   = 0;
-    ws->grabWindow = NULL;
+    if (w == ws->grabWindow)
+    {
+	ws->grabMask   = 0;
+	ws->grabWindow = NULL;
+    }
 
     if (ww->grabbed)
     {
@@ -2786,11 +2603,11 @@ wobblyWindowUngrabNotify (CompWindow *w)
 
 
 static Bool
-wobblyPaintScreen (CompScreen		   *s,
+wobblyPaintOutput (CompScreen		   *s,
 		   const ScreenPaintAttrib *sAttrib,
 		   const CompTransform	   *transform,
 		   Region		   region,
-		   int			   output,
+		   CompOutput		   *output,
 		   unsigned int		   mask)
 {
     Bool status;
@@ -2800,15 +2617,16 @@ wobblyPaintScreen (CompScreen		   *s,
     if (ws->wobblyWindows)
 	mask |= PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK;
 
-    UNWRAP (ws, s, paintScreen);
-    status = (*s->paintScreen) (s, sAttrib, transform, region, output, mask);
-    WRAP (ws, s, paintScreen, wobblyPaintScreen);
+    UNWRAP (ws, s, paintOutput);
+    status = (*s->paintOutput) (s, sAttrib, transform, region, output, mask);
+    WRAP (ws, s, paintOutput, wobblyPaintOutput);
 
     return status;
 }
 
 static CompOption *
-wobblyGetDisplayOptions (CompDisplay *display,
+wobblyGetDisplayOptions (CompPlugin  *plugin,
+			 CompDisplay *display,
 			 int	     *count)
 {
     WOBBLY_DISPLAY (display);
@@ -2818,8 +2636,9 @@ wobblyGetDisplayOptions (CompDisplay *display,
 }
 
 static Bool
-wobblySetDisplayOption (CompDisplay     *display,
-			char	        *name,
+wobblySetDisplayOption (CompPlugin      *plugin,
+			CompDisplay     *display,
+			const char	*name,
 			CompOptionValue *value)
 {
     CompOption *o;
@@ -2832,67 +2651,26 @@ wobblySetDisplayOption (CompDisplay     *display,
 	return FALSE;
 
     switch (index) {
-    case WOBBLY_DISPLAY_OPTION_SNAP:
+    case WOBBLY_DISPLAY_OPTION_SNAP_KEY:
 	/* ignore the key */
 	value->action.key.keycode = 0;
 
 	if (compSetActionOption (o, value))
 	    return TRUE;
 	break;
-    case WOBBLY_DISPLAY_OPTION_SNAP_INVERTED:
-	if (compSetBoolOption (o, value))
-	    return TRUE;
-	break;
-    case WOBBLY_DISPLAY_OPTION_SHIVER:
-	if (setDisplayAction (display, o, value))
-	    return TRUE;
     default:
-	break;
+	return compSetDisplayOption (display, o, value);
     }
 
     return FALSE;
 }
 
-static void
-wobblyDisplayInitOptions (WobblyDisplay *wd)
-{
-    CompOption *o;
-
-    o = &wd->opt[WOBBLY_DISPLAY_OPTION_SNAP];
-    o->name			  = "snap";
-    o->shortDesc		  = N_("Snap windows");
-    o->longDesc			  = N_("Toggle window snapping");
-    o->type			  = CompOptionTypeAction;
-    o->value.action.initiate	  = wobblyEnableSnapping;
-    o->value.action.terminate	  = wobblyDisableSnapping;
-    o->value.action.bell	  = FALSE;
-    o->value.action.edgeMask	  = 0;
-    o->value.action.state	  = 0;
-    o->value.action.type	  = CompBindingTypeKey;
-    o->value.action.key.modifiers = WOBBLY_SNAP_MODIFIERS_DEFAULT;
-    o->value.action.key.keycode   = 0;
-
-    o = &wd->opt[WOBBLY_DISPLAY_OPTION_SNAP_INVERTED];
-    o->name	 = "snap_inverted";
-    o->shortDesc = N_("Snap Inverted");
-    o->longDesc	 = N_("Inverted window snapping");
-    o->type	 = CompOptionTypeBool;
-    o->value.b	 = WOBBLY_SNAP_INVERTED_DEFAULT;
-
-    o = &wd->opt[WOBBLY_DISPLAY_OPTION_SHIVER];
-    o->name		      = "shiver";
-    o->shortDesc	      = N_("Shiver");
-    o->longDesc		      = N_("Make window shiver");
-    o->type		      = CompOptionTypeAction;
-    o->value.action.initiate  = wobblyShiver;
-    o->value.action.terminate = 0;
-    o->value.action.bell      = FALSE;
-    o->value.action.edgeMask  = 0;
-    o->value.action.state     = CompActionStateInitBell;
-    o->value.action.state    |= CompActionStateInitKey;
-    o->value.action.state    |= CompActionStateInitButton;
-    o->value.action.type      = 0;
-}
+static const CompMetadataOptionInfo wobblyDisplayOptionInfo[] = {
+    { "snap_key", "key", "<passive_grab>false</passive_grab>",
+      wobblyEnableSnapping, wobblyDisableSnapping },
+    { "snap_inverted", "bool", 0, 0, 0 },
+    { "shiver", "bell", 0, wobblyShiver, 0 }
+};
 
 static Bool
 wobblyInitDisplay (CompPlugin  *p,
@@ -2900,13 +2678,27 @@ wobblyInitDisplay (CompPlugin  *p,
 {
     WobblyDisplay *wd;
 
+    if (!checkPluginABI ("core", CORE_ABIVERSION))
+	return FALSE;
+
     wd = malloc (sizeof (WobblyDisplay));
     if (!wd)
 	return FALSE;
 
+    if (!compInitDisplayOptionsFromMetadata (d,
+					     &wobblyMetadata,
+					     wobblyDisplayOptionInfo,
+					     wd->opt,
+					     WOBBLY_DISPLAY_OPTION_NUM))
+    {
+	free (wd);
+	return FALSE;
+    }
+
     wd->screenPrivateIndex = allocateScreenPrivateIndex (d);
     if (wd->screenPrivateIndex < 0)
     {
+	compFiniDisplayOptions (d, wd->opt, WOBBLY_DISPLAY_OPTION_NUM);
 	free (wd);
 	return FALSE;
     }
@@ -2915,9 +2707,7 @@ wobblyInitDisplay (CompPlugin  *p,
 
     wd->snapping = FALSE;
 
-    wobblyDisplayInitOptions (wd);
-
-    d->privates[displayPrivateIndex].ptr = wd;
+    d->base.privates[displayPrivateIndex].ptr = wd;
 
     return TRUE;
 }
@@ -2932,6 +2722,8 @@ wobblyFiniDisplay (CompPlugin  *p,
 
     UNWRAP (wd, d, handleEvent);
 
+    compFiniDisplayOptions (d, wd->opt, WOBBLY_DISPLAY_OPTION_NUM);
+
     free (wd);
 }
 
@@ -2947,36 +2739,33 @@ wobblyInitScreen (CompPlugin *p,
     if (!ws)
 	return FALSE;
 
+    if (!compInitScreenOptionsFromMetadata (s,
+					    &wobblyMetadata,
+					    wobblyScreenOptionInfo,
+					    ws->opt,
+					    WOBBLY_SCREEN_OPTION_NUM))
+    {
+	free (ws);
+	return FALSE;
+    }
+
     ws->windowPrivateIndex = allocateWindowPrivateIndex (s);
     if (ws->windowPrivateIndex < 0)
     {
+	compFiniScreenOptions (s, ws->opt, WOBBLY_SCREEN_OPTION_NUM);
 	free (ws);
 	return FALSE;
     }
 
     ws->wobblyWindows = FALSE;
 
-    ws->mapEffect   = WobblyEffectNone;
-    ws->focusEffect = WobblyEffectNone;
-
     ws->grabMask   = 0;
     ws->grabWindow = NULL;
     ws->moveWindow = FALSE;
 
-    wobblyScreenInitOptions (ws, s->display->display);
-
-    matchUpdate (s->display,
-		 &ws->opt[WOBBLY_SCREEN_OPTION_MAP_WINDOW_MATCH].value.match);
-    matchUpdate (s->display,
-		 &ws->opt[WOBBLY_SCREEN_OPTION_FOCUS_WINDOW_MATCH].value.match);
-    matchUpdate (s->display,
-		 &ws->opt[WOBBLY_SCREEN_OPTION_MOVE_WINDOW_MATCH].value.match);
-    matchUpdate (s->display,
-		 &ws->opt[WOBBLY_SCREEN_OPTION_GRAB_WINDOW_MATCH].value.match);
-
     WRAP (ws, s, preparePaintScreen, wobblyPreparePaintScreen);
     WRAP (ws, s, donePaintScreen, wobblyDonePaintScreen);
-    WRAP (ws, s, paintScreen, wobblyPaintScreen);
+    WRAP (ws, s, paintOutput, wobblyPaintOutput);
     WRAP (ws, s, paintWindow, wobblyPaintWindow);
     WRAP (ws, s, damageWindowRect, wobblyDamageWindowRect);
     WRAP (ws, s, addWindowGeometry, wobblyAddWindowGeometry);
@@ -2985,7 +2774,7 @@ wobblyInitScreen (CompPlugin *p,
     WRAP (ws, s, windowGrabNotify, wobblyWindowGrabNotify);
     WRAP (ws, s, windowUngrabNotify, wobblyWindowUngrabNotify);
 
-    s->privates[wd->screenPrivateIndex].ptr = ws;
+    s->base.privates[wd->screenPrivateIndex].ptr = ws;
 
     return TRUE;
 }
@@ -2998,17 +2787,9 @@ wobblyFiniScreen (CompPlugin *p,
 
     freeWindowPrivateIndex (s, ws->windowPrivateIndex);
 
-    free (ws->opt[WOBBLY_SCREEN_OPTION_MAP_EFFECT].value.s);
-    free (ws->opt[WOBBLY_SCREEN_OPTION_FOCUS_EFFECT].value.s);
-
-    matchFini (&ws->opt[WOBBLY_SCREEN_OPTION_MAP_WINDOW_MATCH].value.match);
-    matchFini (&ws->opt[WOBBLY_SCREEN_OPTION_FOCUS_WINDOW_MATCH].value.match);
-    matchFini (&ws->opt[WOBBLY_SCREEN_OPTION_MOVE_WINDOW_MATCH].value.match);
-    matchFini (&ws->opt[WOBBLY_SCREEN_OPTION_GRAB_WINDOW_MATCH].value.match);
-
     UNWRAP (ws, s, preparePaintScreen);
     UNWRAP (ws, s, donePaintScreen);
-    UNWRAP (ws, s, paintScreen);
+    UNWRAP (ws, s, paintOutput);
     UNWRAP (ws, s, paintWindow);
     UNWRAP (ws, s, damageWindowRect);
     UNWRAP (ws, s, addWindowGeometry);
@@ -3017,6 +2798,8 @@ wobblyFiniScreen (CompPlugin *p,
     UNWRAP (ws, s, windowGrabNotify);
     UNWRAP (ws, s, windowUngrabNotify);
 
+    compFiniScreenOptions (s, ws->opt, WOBBLY_SCREEN_OPTION_NUM);
+
     free (ws);
 }
 
@@ -3037,7 +2820,7 @@ wobblyInitWindow (CompPlugin *p,
     ww->grabbed = FALSE;
     ww->state   = w->state;
 
-    w->privates[ws->windowPrivateIndex].ptr = ww;
+    w->base.privates[ws->windowPrivateIndex].ptr = ww;
 
     if (w->mapNum && ws->opt[WOBBLY_SCREEN_OPTION_MAXIMIZE_EFFECT].value.b)
     {
@@ -3053,6 +2836,13 @@ wobblyFiniWindow (CompPlugin *p,
 		  CompWindow *w)
 {
     WOBBLY_WINDOW (w);
+    WOBBLY_SCREEN (w->screen);
+
+    if (ws->grabWindow == w)
+    {
+	ws->grabWindow = NULL;
+	ws->grabMask   = 0;
+    }
 
     if (ww->model)
     {
@@ -3063,12 +2853,84 @@ wobblyFiniWindow (CompPlugin *p,
     free (ww);
 }
 
+static CompBool
+wobblyInitObject (CompPlugin *p,
+		 CompObject *o)
+{
+    static InitPluginObjectProc dispTab[] = {
+	(InitPluginObjectProc) 0, /* InitCore */
+	(InitPluginObjectProc) wobblyInitDisplay,
+	(InitPluginObjectProc) wobblyInitScreen,
+	(InitPluginObjectProc) wobblyInitWindow
+    };
+
+    RETURN_DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), TRUE, (p, o));
+}
+
+static void
+wobblyFiniObject (CompPlugin *p,
+		 CompObject *o)
+{
+    static FiniPluginObjectProc dispTab[] = {
+	(FiniPluginObjectProc) 0, /* FiniCore */
+	(FiniPluginObjectProc) wobblyFiniDisplay,
+	(FiniPluginObjectProc) wobblyFiniScreen,
+	(FiniPluginObjectProc) wobblyFiniWindow
+    };
+
+    DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), (p, o));
+}
+
+static CompOption *
+wobblyGetObjectOptions (CompPlugin *plugin,
+		       CompObject *object,
+		       int	  *count)
+{
+    static GetPluginObjectOptionsProc dispTab[] = {
+	(GetPluginObjectOptionsProc) 0, /* GetCoreOptions */
+	(GetPluginObjectOptionsProc) wobblyGetDisplayOptions,
+	(GetPluginObjectOptionsProc) wobblyGetScreenOptions
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab),
+		     (void *) (*count = 0), (plugin, object, count));
+}
+
+static CompBool
+wobblySetObjectOption (CompPlugin      *plugin,
+		      CompObject      *object,
+		      const char      *name,
+		      CompOptionValue *value)
+{
+    static SetPluginObjectOptionProc dispTab[] = {
+	(SetPluginObjectOptionProc) 0, /* SetCoreOption */
+	(SetPluginObjectOptionProc) wobblySetDisplayOption,
+	(SetPluginObjectOptionProc) wobblySetScreenOption
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab), FALSE,
+		     (plugin, object, name, value));
+}
+
 static Bool
 wobblyInit (CompPlugin *p)
 {
+    if (!compInitPluginMetadataFromInfo (&wobblyMetadata,
+					 p->vTable->name,
+					 wobblyDisplayOptionInfo,
+					 WOBBLY_DISPLAY_OPTION_NUM,
+					 wobblyScreenOptionInfo,
+					 WOBBLY_SCREEN_OPTION_NUM))
+	return FALSE;
+
     displayPrivateIndex = allocateDisplayPrivateIndex ();
     if (displayPrivateIndex < 0)
+    {
+	compFiniMetadata (&wobblyMetadata);
 	return FALSE;
+    }
+
+    compAddMetadataFromFile (&wobblyMetadata, p->vTable->name);
 
     return TRUE;
 }
@@ -3076,48 +2938,29 @@ wobblyInit (CompPlugin *p)
 static void
 wobblyFini (CompPlugin *p)
 {
-    if (displayPrivateIndex >= 0)
-	freeDisplayPrivateIndex (displayPrivateIndex);
+    freeDisplayPrivateIndex (displayPrivateIndex);
+    compFiniMetadata (&wobblyMetadata);
 }
 
-static int
-wobblyGetVersion (CompPlugin *plugin,
-		  int	     version)
+static CompMetadata *
+wobblyGetMetadata (CompPlugin *plugin)
 {
-    return ABIVERSION;
+    return &wobblyMetadata;
 }
 
-CompPluginDep wobblyDeps[] = {
-    { CompPluginRuleBefore, "fade" },
-    { CompPluginRuleBefore, "cube" },
-    { CompPluginRuleBefore, "scale" }
-};
-
 CompPluginVTable wobblyVTable = {
     "wobbly",
-    N_("Wobbly Windows"),
-    N_("Use spring model for wobbly window effect"),
-    wobblyGetVersion,
+    wobblyGetMetadata,
     wobblyInit,
     wobblyFini,
-    wobblyInitDisplay,
-    wobblyFiniDisplay,
-    wobblyInitScreen,
-    wobblyFiniScreen,
-    wobblyInitWindow,
-    wobblyFiniWindow,
-    wobblyGetDisplayOptions,
-    wobblySetDisplayOption,
-    wobblyGetScreenOptions,
-    wobblySetScreenOption,
-    wobblyDeps,
-    sizeof (wobblyDeps) / sizeof (wobblyDeps[0]),
-    0, /* Features */
-    0  /* nFeatures */
+    wobblyInitObject,
+    wobblyFiniObject,
+    wobblyGetObjectOptions,
+    wobblySetObjectOption
 };
 
 CompPluginVTable *
-getCompPluginInfo (void)
+getCompPluginInfo20070830 (void)
 {
     return &wobblyVTable;
 }
